"""GluonCV-Torch video action recognition."""
from __future__ import absolute_import
from .actionrec_resnetv1b import *
from .i3d_resnet import *
from .i3d_slow import *
from .slowfast import *
from .r2plus1dv1 import *
from .r2plus1dv2 import *
from .tpn import *
from .ircsnv2 import *



"""Utils for auto YOLO estimator"""
import os

from mxnet import gluon

from ....data import MixupDetection
from ....data.batchify import Tuple, Stack, Pad
from ....data.dataloader import RandomTransformDataLoader
from ....data.transforms.presets.yolo import YOLO3DefaultTrainTransform
from ....data.transforms.presets.yolo import YOLO3DefaultValTransform
from .... import data as gdata
from ....utils.metrics.voc_detection import VOC07MApMetric
from ....utils.metrics.coco_detection import COCODetectionMetric


def _get_dataset(dataset, args):
    if dataset.lower() == 'voc':
        train_dataset = gdata.VOCDetection(
            splits=[(2007, 'trainval'), (2012, 'trainval')])
        val_dataset = gdata.VOCDetection(
            splits=[(2007, 'test')])
        val_metric = VOC07MApMetric(iou_thresh=0.5, class_names=val_dataset.classes)
    elif dataset.lower() == 'voc_tiny':
        # need to download the dataset and specify the path to store the dataset in
        # root = os.path.expanduser('~/.mxnet/datasets/')
        # filename_zip = ag.download('https://autogluon.s3.amazonaws.com/datasets/tiny_motorbike.zip', path=root)
        # filename = ag.unzip(filename_zip, root=root)
        # data_root = os.path.join(root, filename)
        train_dataset = gdata.CustomVOCDetectionBase(classes=('motorbike',), root=args.dataset_root + 'tiny_motorbike',
                                                     splits=[('', 'trainval')])
        val_dataset = gdata.CustomVOCDetectionBase(classes=('motorbike',), root=args.dataset_root + 'tiny_motorbike',
                                                   splits=[('', 'test')])
        val_metric = VOC07MApMetric(iou_thresh=0.5, class_names=val_dataset.classes)
    elif dataset.lower() == 'coco':
        train_dataset = gdata.COCODetection(splits='instances_train2017', use_crowd=False)
        val_dataset = gdata.COCODetection(splits='instances_val2017', skip_empty=False)
        val_metric = COCODetectionMetric(
            val_dataset, os.path.join(args.logdir, args.save_prefix + '_eval'), cleanup=True,
            data_shape=(args.yolo3.data_shape, args.yolo3.data_shape))
    else:
        raise NotImplementedError('Dataset: {} not implemented.'.format(dataset))
    if args.train.num_samples < 0:
        args.train.num_samples = len(train_dataset)
    if args.train.mixup:
        train_dataset = MixupDetection(train_dataset)
    return train_dataset, val_dataset, val_metric

def _get_dataloader(net, train_dataset, val_dataset, data_shape, batch_size, num_workers, args):
    """Get dataloader."""
    width, height = data_shape, data_shape
    # stack image, all targets generated
    batchify_fn = Tuple(*([Stack() for _ in range(6)] + [Pad(axis=0, pad_val=-1) for _ in range(1)]))
    if args.yolo3.no_random_shape:
        train_loader = gluon.data.DataLoader(
            train_dataset.transform(YOLO3DefaultTrainTransform(width, height, net, mixup=args.train.mixup)),
            batch_size, True, batchify_fn=batchify_fn, last_batch='rollover', num_workers=num_workers)
    else:
        transform_fns = [YOLO3DefaultTrainTransform(x * 32, x * 32, net, mixup=args.train.mixup) for x in range(10, 20)]
        train_loader = RandomTransformDataLoader(
            transform_fns, train_dataset, batch_size=batch_size, interval=10, last_batch='rollover',
            shuffle=True, batchify_fn=batchify_fn, num_workers=num_workers)
    val_batchify_fn = Tuple(Stack(), Pad(pad_val=-1))
    val_loader = gluon.data.DataLoader(
        val_dataset.transform(YOLO3DefaultValTransform(width, height)),
        batch_size, False, batchify_fn=val_batchify_fn, last_batch='keep', num_workers=num_workers)
    train_eval_loader = gluon.data.DataLoader(
        train_dataset.transform(YOLO3DefaultValTransform(width, height)),
        batch_size, False, batchify_fn=val_batchify_fn, last_batch='keep', num_workers=num_workers)
    return train_loader, val_loader, train_eval_loader

def _save_params(net, best_map, current_map, epoch, save_interval, prefix):
    current_map = float(current_map)
    if current_map > best_map[0]:
        best_map[0] = current_map
        net.save_parameters('{:s}_{:04d}_{:.4f}_best.params'.format(prefix, epoch, current_map))
        with open(prefix+'_best_map.log', 'a') as log_file:
            log_file.write('{:04d}:\t{:.4f}\n'.format(epoch, current_map))
    if save_interval and epoch % save_interval == 0:
        net.save_parameters('{:s}_{:04d}_{:.4f}.params'.format(prefix, epoch, current_map))



"""SSD Estimator."""
# pylint: disable=logging-format-interpolation,abstract-method
import os
import time
import warnings

import pandas as pd
import numpy as np
import mxnet as mx
from mxnet import nd
from mxnet import gluon
from mxnet import autograd
from mxnet.contrib import amp

from .... import utils as gutils
from ....data.batchify import Tuple, Stack, Pad
from ....utils.metrics.voc_detection import VOC07MApMetric, VOCMApMetric
from ....model_zoo import get_model
from ....model_zoo import custom_ssd
from ....data.transforms.presets.ssd import SSDDefaultValTransform
from ....data.transforms.presets.ssd import load_test, transform_test
from ....loss import SSDMultiBoxLoss
from .utils import _get_dataloader, _get_dali_dataloader
from ..base_estimator import BaseEstimator, set_default
from .default import SSDCfg

try:
    import horovod.mxnet as hvd
except ImportError:
    hvd = None

try:
    dali_found = True
except ImportError:
    dali_found = False

__all__ = ['SSDEstimator']


@set_default(SSDCfg())
class SSDEstimator(BaseEstimator):
    """Estimator implementation for SSD.

    Parameters
    ----------
    config : dict
        Config in nested dict.
    logger : logging.Logger, default is None
        Optional logger for this estimator, can be `None` when default setting is used.
    reporter : callable, default is None
        If set, use reporter callback to report the metrics of the current estimator.

    Attributes
    ----------
    _logger : logging.Logger
        The customized/default logger for this estimator.
    _logdir : str
        The temporary dir for logs.
    _cfg : autocfg.dataclass
        The configurations.
    """
    def __init__(self, config, logger=None, reporter=None):
        super(SSDEstimator, self).__init__(config, logger, reporter, name=self.__class__.__name__)

        if self._cfg.ssd.amp:
            amp.init()
        if self._cfg.horovod:
            hvd.init()

    def _fit(self, train_data, val_data):
        """Fit SSD model."""
        self._best_map = 0
        self.epoch = 0
        self._time_elapsed = 0
        if max(self._cfg.train.start_epoch, self.epoch) >= self._cfg.train.epochs:
            return {'time', self._time_elapsed}
        self.net.collect_params().reset_ctx(self.ctx)
        self._init_trainer()
        return self._resume_fit(train_data, val_data)

    def _resume_fit(self, train_data, val_data):
        if max(self._cfg.train.start_epoch, self.epoch) >= self._cfg.train.epochs:
            return {'time', self._time_elapsed}
        if not self.classes or not self.num_class:
            raise ValueError('Unable to determine classes of dataset')

        # dataset
        devices = [int(i) for i in self._cfg.gpus]
        train_dataset = train_data.to_mxnet()
        val_dataset = val_data.to_mxnet()

        # dataloader
        if self._cfg.train.dali:
            if not dali_found:
                raise SystemExit("DALI not found, please check if you installed it correctly.")
            train_loader, val_loader = _get_dali_dataloader(
                self.async_net, train_dataset, val_dataset, self._cfg.ssd.data_shape,
                self._cfg.train.batch_size, self._cfg.num_workers,
                devices, self.ctx[0], self._cfg.horovod)
        else:
            self.batch_size = self._cfg.train.batch_size // hvd.size() \
                if self._cfg.horovod else self._cfg.train.batch_size
            train_loader, val_loader, train_eval_loader = _get_dataloader(
                self.async_net, train_dataset, val_dataset, self._cfg.ssd.data_shape,
                self.batch_size, self._cfg.num_workers)

        return self._train_loop(train_loader, val_loader, train_eval_loader)

    def _train_loop(self, train_data, val_data, train_eval_data):
        # fix seed for mxnet, numpy and python builtin random generator.
        gutils.random.seed(self._cfg.train.seed)
        # loss and metric
        mbox_loss = SSDMultiBoxLoss()
        ce_metric = mx.metric.Loss('CrossEntropy')
        smoothl1_metric = mx.metric.Loss('SmoothL1')

        # lr decay policy
        lr_decay = float(self._cfg.train.lr_decay)
        lr_steps = sorted([float(ls) for ls in self._cfg.train.lr_decay_epoch])

        self._logger.info('Start training from [Epoch %d]', max(self._cfg.train.start_epoch, self.epoch))

        self.net.collect_params().reset_ctx(self.ctx)
        for self.epoch in range(max(self._cfg.train.start_epoch, self.epoch), self._cfg.train.epochs):
            epoch = self.epoch
            while lr_steps and epoch >= lr_steps[0]:
                new_lr = self.trainer.learning_rate * lr_decay
                lr_steps.pop(0)
                self.trainer.set_learning_rate(new_lr)
                self._logger.info("[Epoch {}] Set learning rate to {}".format(epoch, new_lr))
            ce_metric.reset()
            smoothl1_metric.reset()
            tic = time.time()
            btic = time.time()
            self.net.hybridize(static_alloc=True, static_shape=True)

            for i, batch in enumerate(train_data):
                if self._cfg.train.dali:
                    # dali iterator returns a mxnet.io.DataBatch
                    data = [d.data[0] for d in batch]
                    box_targets = [d.label[0] for d in batch]
                    cls_targets = [nd.cast(d.label[1], dtype='float32') for d in batch]
                else:
                    data = gluon.utils.split_and_load(batch[0], ctx_list=self.ctx, batch_axis=0, even_split=False)
                    cls_targets = gluon.utils.split_and_load(batch[1], ctx_list=self.ctx,
                                                             batch_axis=0, even_split=False)
                    box_targets = gluon.utils.split_and_load(batch[2], ctx_list=self.ctx,
                                                             batch_axis=0, even_split=False)

                with autograd.record():
                    cls_preds = []
                    box_preds = []
                    for x in data:
                        cls_pred, box_pred, _ = self.net(x)
                        cls_preds.append(cls_pred)
                        box_preds.append(box_pred)
                    sum_loss, cls_loss, box_loss = mbox_loss(
                        cls_preds, box_preds, cls_targets, box_targets)
                    if self._cfg.ssd.amp:
                        with amp.scale_loss(sum_loss, self.trainer) as scaled_loss:
                            autograd.backward(scaled_loss)
                    else:
                        autograd.backward(sum_loss)
                # since we have already normalized the loss, we don't want to normalize
                # by batch-size anymore
                self.trainer.step(1)

                if not self._cfg.horovod or hvd.rank() == 0:
                    local_batch_size = int(self._cfg.train.batch_size // (hvd.size() if self._cfg.horovod else 1))
                    ce_metric.update(0, [l * local_batch_size for l in cls_loss])
                    smoothl1_metric.update(0, [l * local_batch_size for l in box_loss])
                    if self._cfg.train.log_interval and not (i + 1) % self._cfg.train.log_interval:
                        name1, loss1 = ce_metric.get()
                        name2, loss2 = smoothl1_metric.get()
                        self._logger.info('[Epoch %d][Batch %d], Speed: %f samples/sec, %s=%f, %s=%f',
                                          epoch, i, self._cfg.train.batch_size/(time.time()-btic),
                                          name1, loss1, name2, loss2)
                    btic = time.time()

            if not self._cfg.horovod or hvd.rank() == 0:
                name1, loss1 = ce_metric.get()
                name2, loss2 = smoothl1_metric.get()
                self._logger.info('[Epoch %d] Training cost: %f, %s=%f, %s=%f',
                                  epoch, (time.time()-tic), name1, loss1, name2, loss2)
                if (epoch % self._cfg.valid.val_interval == 0) or \
                    (self._cfg.save_interval and epoch % self._cfg.save_interval == 0):
                    # consider reduce the frequency of validation to save time
                    map_name, mean_ap = self._evaluate(val_data)
                    val_msg = '\n'.join(['{}={}'.format(k, v) for k, v in zip(map_name, mean_ap)])
                    self._logger.info('[Epoch %d] Validation: \n%s', epoch, str(val_msg))
                    current_map = float(mean_ap[-1])
                    if current_map > self._best_map:
                        cp_name = os.path.join(self._logdir, 'best_checkpoint.pkl')
                        self._logger.info('[Epoch %d] Current best map: %f vs previous %f, saved to %s',
                                          self.epoch, current_map, self._best_map, cp_name)
                        self.save(cp_name)
                        self._best_map = current_map
                if self._reporter:
                    self._reporter(epoch=epoch, map_reward=current_map)
            self._time_elapsed += time.time() - btic
        # map on train data
        map_name, mean_ap = self._evaluate(train_eval_data)
        return {'train_map': float(mean_ap[-1]), 'valid_map': self._best_map, 'time': self._time_elapsed}

    def _evaluate(self, val_data):
        """Evaluate on validation dataset."""
        if not isinstance(val_data, gluon.data.DataLoader):
            if hasattr(val_data, 'to_mxnet'):
                val_data = val_data.to_mxnet()
            val_batchify_fn = Tuple(Stack(), Pad(pad_val=-1))
            val_data = gluon.data.DataLoader(
                val_data.transform(SSDDefaultValTransform(self._cfg.ssd.data_shape, self._cfg.ssd.data_shape)),
                self._cfg.valid.batch_size, False, batchify_fn=val_batchify_fn, last_batch='keep',
                num_workers=self._cfg.num_workers)
        if self._cfg.valid.metric == 'voc07':
            eval_metric = VOC07MApMetric(iou_thresh=self._cfg.valid.iou_thresh, class_names=self.classes)
        elif self._cfg.valid.metric == 'voc':
            eval_metric = VOCMApMetric(iou_thresh=self._cfg.valid.iou_thresh, class_names=self.classes)
        else:
            raise ValueError(f'Invalid metric type: {self._cfg.valid.metric}')
        self.net.set_nms(nms_thresh=0.45, nms_topk=400)
        self.net.collect_params().reset_ctx(self.ctx)
        self.net.hybridize(static_alloc=True, static_shape=True)
        for batch in val_data:
            data = gluon.utils.split_and_load(batch[0], ctx_list=self.ctx, batch_axis=0, even_split=False)
            label = gluon.utils.split_and_load(batch[1], ctx_list=self.ctx, batch_axis=0, even_split=False)
            det_bboxes = []
            det_ids = []
            det_scores = []
            gt_bboxes = []
            gt_ids = []
            gt_difficults = []
            for x, y in zip(data, label):
                # get prediction results
                ids, scores, bboxes = self.net(x)
                det_ids.append(ids)
                det_scores.append(scores)
                # clip to image size
                det_bboxes.append(bboxes.clip(0, batch[0].shape[2]))
                # split ground truths
                gt_ids.append(y.slice_axis(axis=-1, begin=4, end=5))
                gt_bboxes.append(y.slice_axis(axis=-1, begin=0, end=4))
                gt_difficults.append(y.slice_axis(axis=-1, begin=5, end=6) if y.shape[-1] > 5 else None)
            # update metric
            eval_metric.update(det_bboxes, det_ids, det_scores, gt_bboxes, gt_ids, gt_difficults)
        return eval_metric.get()

    def _predict(self, x):
        """Predict an individual example."""
        short_size = int(self._cfg.ssd.data_shape)
        if isinstance(x, str):
            x = load_test(x, short=short_size, max_size=1024)[0]
        elif isinstance(x, mx.nd.NDArray):
            x = transform_test(x, short=short_size, max_size=1024)[0]
        elif isinstance(x, pd.DataFrame):
            assert 'image' in x.columns, "Expect column `image` for input images"
            def _predict_merge(x):
                y = self._predict(x)
                y['image'] = x
                return y
            return pd.concat([_predict_merge(xx) for xx in x['image']]).reset_index(drop=True)
        else:
            raise ValueError('Input is not supported: {}'.format(type(x)))
        height, width = x.shape[2:4]
        x = x.as_in_context(self.ctx[0])
        ids, scores, bboxes = [xx[0].asnumpy() for xx in self.net(x)]
        bboxes[:, (0, 2)] /= width
        bboxes[:, (1, 3)] /= height
        bboxes = np.clip(bboxes, 0.0, 1.0).tolist()
        df = pd.DataFrame({'predict_class': [self.classes[int(id)] for id in ids], 'predict_score': scores.flatten(),
                           'predict_rois': [{'xmin': bbox[0], 'ymin': bbox[1], 'xmax': bbox[2], 'ymax': bbox[3]} \
                                for bbox in bboxes]})
        # filter out invalid (scores < 0) rows
        valid_df = df[df['predict_score'] > 0].reset_index(drop=True)
        return valid_df

    def _init_network(self):
        if not self.num_class:
            raise ValueError('Unable to create network when `num_class` is unknown. \
                It should be inferred from dataset or resumed from saved states.')
        assert len(self.classes) == self.num_class

        # training contexts
        if self._cfg.horovod:
            self.ctx = [mx.gpu(hvd.local_rank())]
        else:
            ctx = [mx.gpu(int(i)) for i in self._cfg.gpus]
            self.ctx = ctx if ctx else [mx.cpu()]

        if self._cfg.ssd.transfer is not None:
            assert isinstance(self._cfg.ssd.transfer, str)
            data_shape = int(self._cfg.ssd.transfer.split('_')[1])
            self._cfg.ssd.data_shape = data_shape
            self._logger.info(
                f'Using transfer learning from {self._cfg.ssd.transfer}, the other network parameters are ignored.')
            if self._cfg.ssd.syncbn and len(self.ctx) > 1:
                self.net = get_model(self._cfg.ssd.transfer, pretrained=True, norm_layer=gluon.contrib.nn.SyncBatchNorm,
                                     norm_kwargs={'num_devices': len(self.ctx)})
                self.async_net = get_model(self._cfg.ssd.transfer, pretrained=True)  # used by cpu worker
                self.net.reset_class(self.classes,
                                     reuse_weights=[cname for cname in self.classes if cname in self.net.classes])
            else:
                self.net = get_model(self._cfg.ssd.transfer, pretrained=True, norm_layer=gluon.nn.BatchNorm)
                self.async_net = get_model(self._cfg.ssd.transfer, pretrained=True, norm_layer=gluon.nn.BatchNorm)
                self.net.reset_class(self.classes,
                                     reuse_weights=[cname for cname in self.classes if cname in self.net.classes])
        # elif self._cfg.ssd.custom_model:
        else:
            if self._cfg.ssd.syncbn and len(self.ctx) > 1:
                self.net = custom_ssd(base_network_name=self._cfg.ssd.backbone,
                                      base_size=self._cfg.ssd.data_shape,
                                      filters=self._cfg.ssd.filters,
                                      sizes=self._cfg.ssd.sizes,
                                      ratios=self._cfg.ssd.ratios,
                                      steps=self._cfg.ssd.steps,
                                      classes=self.classes,
                                      dataset='auto',
                                      pretrained_base=True,
                                      norm_layer=gluon.contrib.nn.SyncBatchNorm,
                                      norm_kwargs={'num_devices': len(self.ctx)})
                self.async_net = custom_ssd(base_network_name=self._cfg.ssd.backbone,
                                            base_size=self._cfg.ssd.data_shape,
                                            filters=self._cfg.ssd.filters,
                                            sizes=self._cfg.ssd.sizes,
                                            ratios=self._cfg.ssd.ratios,
                                            steps=self._cfg.ssd.steps,
                                            classes=self.classes,
                                            dataset='auto',
                                            pretrained_base=False)
            else:
                self.net = custom_ssd(base_network_name=self._cfg.ssd.backbone,
                                      base_size=self._cfg.ssd.data_shape,
                                      filters=self._cfg.ssd.filters,
                                      sizes=self._cfg.ssd.sizes,
                                      ratios=self._cfg.ssd.ratios,
                                      steps=self._cfg.ssd.steps,
                                      classes=self.classes,
                                      dataset=self._cfg.dataset,
                                      pretrained_base=True,
                                      norm_layer=gluon.nn.BatchNorm)
                self.async_net = self.net

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            self.net.initialize()
            self.async_net.initialize()
            # needed for net to be first gpu when using AMP
            self.net.collect_params().reset_ctx(self.ctx)

    def _init_trainer(self):
        if self._cfg.horovod:
            hvd.broadcast_parameters(self.net.collect_params(), root_rank=0)
            self.trainer = hvd.DistributedTrainer(
                self.net.collect_params(), 'sgd',
                {'learning_rate': self._cfg.train.lr, 'wd': self._cfg.train.wd,
                 'momentum': self._cfg.train.momentum})
        else:
            self.trainer = gluon.Trainer(
                self.net.collect_params(), 'sgd',
                {'learning_rate': self._cfg.train.lr, 'wd': self._cfg.train.wd,
                 'momentum': self._cfg.train.momentum},
                update_on_kvstore=(False if self._cfg.ssd.amp else None))

        if self._cfg.ssd.amp:
            amp.init_trainer(self.trainer)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# coding: utf-8
# pylint: disable= arguments-differ
# pylint: disable= missing-docstring
"Addtional image transforms."

import random
import math
import numpy as np
from mxnet import image, nd
from mxnet.gluon import Block


__all__ = ['RandomCrop', 'RandomErasing']


class RandomCrop(Block):
    """Randomly crop `src` with `size` (width, height).
    Padding is optional.
    Upsample result if `src` is smaller than `size`.

    Parameters
    ----------
    size : int or tuple of (W, H)
        Size of the final output.
    pad: int or tuple
        if int, size of the zero-padding
        if tuple, number of values padded to the edges of each axis.
            ((before_1, after_1), ... (before_N, after_N)) unique pad widths for each axis.
            ((before, after),) yields same before and after pad for each axis.
            (pad,) or int is a shortcut for before = after = pad width for all axes.
    interpolation : int
        Interpolation method for resizing. By default uses bilinear
        interpolation. See OpenCV's resize function for available choices.


    Inputs:
        - **data**: input tensor with (Hi x Wi x C) shape.
    Outputs:
        - **out**: output tensor with (size[0] x size[1] x C) or (size x size x C) shape.
    """

    def __init__(self, size, pad=None, interpolation=2):
        super(RandomCrop, self).__init__()
        numeric_types = (float, int, np.generic)
        if isinstance(size, numeric_types):
            size = (size, size)
        self._args = (size, interpolation)
        self.pad = ((pad, pad), (pad, pad), (0, 0)) if isinstance(pad, int) else pad
    def forward(self, x):
        if self.pad:
            return image.random_crop(nd.array(
                np.pad(x.asnumpy(), self.pad, mode='constant', constant_values=0)), *self._args)[0]
        else:
            return image.random_crop(x, *self._args)[0]

class RandomErasing(Block):
    """Randomly erasing the area in `src` between `s_min` and `s_max` with `probability`.
    `ratio` controls the ratio between width and height.
    `mean` means the value in erasing area.

    Parameters
    ----------
    probability : float
        Probability of erasing.
    s_min : float
        Min area to all area.
    s_max : float
        Max area to all area.
    ratio : float
        The ratio between width and height.
    mean : int or tuple of (R, G, B)
        The value in erasing area.


    Inputs:
        - **data**: input tensor with (Hi x Wi x C) shape.
    Outputs:
        - **out**: output tensor with (Hi x Wi x C) shape.
    """

    def __init__(self, probability=0.5, s_min=0.02, s_max=0.4, ratio=0.3,
                 mean=(125.31, 122.96, 113.86)):
        super(RandomErasing, self).__init__()
        self.probability = probability
        self.mean = mean
        self.s_min = s_min
        self.s_max = s_max
        self.ratio = ratio

    def forward(self, x):
        if not isinstance(self.probability, float):
            raise TypeError('Got inappropriate size arg')
        if not isinstance(self.s_min, float):
            raise TypeError('Got inappropriate size arg')
        if not isinstance(self.s_max, float):
            raise TypeError('Got inappropriate size arg')
        if not isinstance(self.ratio, float):
            raise TypeError('Got inappropriate size arg')
        if not isinstance(self.mean, (int, tuple)):
            raise TypeError('Got inappropriate size arg')

        if random.uniform(0, 1) > self.probability:
            return x

        width, height, _ = x.shape
        area = width * height
        target_area = random.uniform(self.s_min, self.s_max) * area
        aspect_ratio = random.uniform(self.ratio, 1/self.ratio)
        w = int(round(math.sqrt(target_area * aspect_ratio)))
        h = int(round(math.sqrt(target_area / aspect_ratio)))
        if w < width and h < height:
            x1 = random.randint(0, width - w)
            y1 = random.randint(0, height - h)
            x[x1:x1+w, y1:y1+h, 0] = self.mean[0]
            x[x1:x1+w, y1:y1+h, 1] = self.mean[1]
            x[x1:x1+w, y1:y1+h, 2] = self.mean[2]
        return x



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# coding: utf-8
# pylint: disable= arguments-differ,unused-argument
"""ResNet V1b series modified for SSD detection"""
from __future__ import absolute_import

import mxnet as mx
from mxnet.gluon import HybridBlock
from mxnet.gluon.nn import HybridSequential, Conv2D, Activation, BatchNorm
from ..model_zoo import get_model


class ResNetV1bSSD(HybridBlock):
    """Single-shot Object Detection Network: https://arxiv.org/abs/1512.02325.
       with resnetv1b base model.

    Parameters
    ----------
    network : string or None
        Name of the base network, must be end with v1b.
    add_filters : list of int
        Number of channels for the appended layers, ignored if `network`is `None`.
    norm_layer : object
        Normalization layer used (default: :class:`mxnet.gluon.nn.BatchNorm`)
        Can be :class:`mxnet.gluon.nn.BatchNorm` or :class:`mxnet.gluon.contrib.nn.SyncBatchNorm`.
        This will only apply to base networks that has `norm_layer` specified, will ignore if the
        base network (e.g. VGG) don't accept this argument.
    norm_kwargs : dict
        Additional `norm_layer` arguments, for example `num_devices=4`
        for :class:`mxnet.gluon.contrib.nn.SyncBatchNorm`.
    use_bn : bool
        Whether to use BatchNorm layer after each attached convolutional layer.
    reduce_ratio : float
        Channel reduce ratio (0, 1) of the transition layer.
    min_depth : int
        Minimum channels for the transition layers.
    """
    def __init__(self, network, add_filters,
                 norm_layer=BatchNorm, norm_kwargs=None,
                 use_bn=False, reduce_ratio=1.0, min_depth=128, **kwargs):
        super(ResNetV1bSSD, self).__init__()
        assert network.endswith('v1b')
        if norm_kwargs is None:
            norm_kwargs = {}
        res = get_model(network, **kwargs)
        weight_init = mx.init.Xavier(rnd_type='gaussian', factor_type='out', magnitude=2)
        with self.name_scope():
            self.stage1 = HybridSequential('stage1')
            for l in ['conv1', 'bn1', 'relu', 'maxpool', 'layer1', 'layer2']:
                self.stage1.add(getattr(res, l))
            self.stage2 = HybridSequential('stage2')
            self.stage2.add(res.layer3)
            # set stride from (2, 2) -> (1, 1) in first conv of layer3
            self.stage2[0][0].conv1._kwargs['stride'] = (1, 1)
            # also the residuel path
            self.stage2[0][0].downsample[0]._kwargs['stride'] = (1, 1)
            self.stage2.add(res.layer4)
            self.more_stages = HybridSequential('more_stages')
            for i, num_filter in enumerate(add_filters):
                stage = HybridSequential('more_stages_' + str(i))
                num_trans = max(min_depth, int(round(num_filter * reduce_ratio)))
                stage.add(Conv2D(channels=num_trans, kernel_size=1, use_bias=not use_bn,
                                 weight_initializer=weight_init))
                if use_bn:
                    stage.add(norm_layer(**norm_kwargs))
                stage.add(Activation('relu'))
                padding = 0 if i == len(add_filters) - 1 else 1
                stage.add(Conv2D(channels=num_filter, kernel_size=3,
                                 strides=2, padding=padding, use_bias=not use_bn,
                                 weight_initializer=weight_init))
                if use_bn:
                    stage.add(norm_layer(**norm_kwargs))
                stage.add(Activation('relu'))
                self.more_stages.add(stage)

    def hybrid_forward(self, F, x):
        y1 = self.stage1(x)
        y2 = self.stage2(y1)
        more_out = [y1, y2]
        out = y2
        for stage in self.more_stages:
            out = stage(out)
            more_out.append(out)
        return more_out

def resnet34_v1b_ssd(**kwargs):
    return ResNetV1bSSD(network='resnet34_v1b', add_filters=[256, 256, 128, 128], **kwargs)



import numpy as np
import PIL

import torch

from .utils import images as imageutils


class ToStackedTensor(object):
    """Converts a list of m (H x W x C) numpy.ndarrays in the range [0, 255]
    or PIL Images to a torch.FloatTensor of shape (m*C x H x W)
    in the range [0, 1.0]
    """

    def __init__(self, channel_nb=3):
        self.channel_nb = channel_nb

    def __call__(self, clip):
        """
        Args:
            clip (list of numpy.ndarray or PIL.Image.Image): clip
            (list of images) to be converted to tensor.
        """
        # Retrieve shape
        if isinstance(clip[0], np.ndarray):
            h, w, ch = clip[0].shape
            assert ch == self.channel_nb, 'got {} channels instead of 3'.format(
                ch)
        elif isinstance(clip[0], PIL.Image.Image):
            w, h = clip[0].size
        else:
            raise TypeError('Expected numpy.ndarray or PIL.Image\
            but got list of {0}'.format(type(clip[0])))

        np_clip = np.zeros([self.channel_nb * len(clip), int(h), int(w)])

        # Convert
        for img_idx, img in enumerate(clip):
            if isinstance(img, np.ndarray):
                pass
            elif isinstance(img, PIL.Image.Image):
                img = np.array(img, copy=False)
            else:
                raise TypeError('Expected numpy.ndarray or PIL.Image\
                but got list of {0}'.format(type(clip[0])))
            img = imageutils.convert_img(img)
            np_clip[img_idx * self.channel_nb:(
                img_idx + 1) * self.channel_nb, :, :] = img
        tensor_clip = torch.from_numpy(np_clip)
        return tensor_clip.float().div(255)



from __future__ import print_function

import numpy as np
import gluoncv as gcv
from mxnet.gluon.nn import BatchNorm

def check_bn_frozen_callback(net, value):
    if isinstance(net, BatchNorm):
        assert value == net._kwargs['use_global_stats']

def test_block_freeze_bn():
    net = gcv.model_zoo.get_model('resnet18_v1')
    gcv.utils.recursive_visit(net, check_bn_frozen_callback, value=False)
    gcv.utils.freeze_bn(net, True)
    gcv.utils.recursive_visit(net, check_bn_frozen_callback, value=True)
    gcv.utils.freeze_bn(net, False)
    gcv.utils.recursive_visit(net, check_bn_frozen_callback, value=False)

if __name__ == '__main__':
    import nose
    nose.runmodule()



"""Faster RCNN Estimator."""
# pylint: disable=logging-not-lazy,abstract-method
import os
import time

import pandas as pd
import numpy as np
import mxnet as mx
from mxnet import gluon

from ....data.batchify import Tuple, Append
from ....data.transforms.presets.rcnn import load_test, transform_test
from ....data.transforms.presets.rcnn import FasterRCNNDefaultTrainTransform, FasterRCNNDefaultValTransform
from ....model_zoo import get_model
from ....model_zoo.rcnn.faster_rcnn.data_parallel import ForwardBackwardTask
from ....nn.bbox import BBoxClipToImage
from ....utils.parallel import Parallel
from ....utils.metrics.rcnn import RPNAccMetric, RPNL1LossMetric, RCNNAccMetric, RCNNL1LossMetric
from ....utils.metrics.voc_detection import VOC07MApMetric, VOCMApMetric
from ..base_estimator import BaseEstimator, set_default
from .utils import _get_lr_at_iter, _get_dataloader, _split_and_load

try:
    import horovod.mxnet as hvd
except ImportError:
    hvd = None

from .default import FasterRCNNCfg

__all__ = ['FasterRCNNEstimator']


@set_default(FasterRCNNCfg())
class FasterRCNNEstimator(BaseEstimator):
    """Estimator implementation for Faster-RCNN.

    Parameters
    ----------
    config : dict
        Config in nested dict.
    logger : logging.Logger, default is None
        Optional logger for this estimator, can be `None` when default setting is used.
    reporter : callable, default is None
        If set, use reporter callback to report the metrics of the current estimator.

    Attributes
    ----------
    _logger : logging.Logger
        The customized/default logger for this estimator.
    _logdir : str
        The temporary dir for logs.
    _cfg : autocfg.dataclass
        The configurations.
    """
    def __init__(self, config, logger=None, reporter=None):
        super(FasterRCNNEstimator, self).__init__(config, logger, reporter)

    def _fit(self, train_data, val_data):
        """Fit Faster R-CNN model."""
        self._best_map = 0
        self.epoch = 0
        self._time_elapsed = 0
        if max(self._cfg.train.start_epoch, self.epoch) >= self._cfg.train.epochs:
            return {'time', self._time_elapsed}
        self.net.collect_params().setattr('grad_req', 'null')
        self.net.collect_train_params().setattr('grad_req', 'write')
        self._init_trainer()
        return self._resume_fit(train_data, val_data)

    def _resume_fit(self, train_data, val_data):
        if max(self._cfg.train.start_epoch, self.epoch) >= self._cfg.train.epochs:
            return {'time', self._time_elapsed}
        if not self.classes or not self.num_class:
            raise ValueError('Unable to determine classes of dataset')

        # dataset
        train_dataset = train_data.to_mxnet()
        val_dataset = val_data.to_mxnet()

        # dataloader
        self.batch_size = self._cfg.train.batch_size // min(1, self.num_gpus) \
            if self._cfg.horovod else self._cfg.train.batch_size
        if not self._cfg.horovod:
            if self._cfg.train.batch_size == 1 and self.num_gpus > 1:
                self.batch_size *= self.num_gpus
        train_loader, val_loader, train_eval_loader = _get_dataloader(
            self.net, train_dataset, val_dataset, FasterRCNNDefaultTrainTransform,
            FasterRCNNDefaultValTransform, self.batch_size, len(self.ctx), self._cfg)

        return self._train_loop(train_loader, val_loader, train_eval_loader)

    def _train_loop(self, train_data, val_data, train_eval_data):
        # loss and metric
        rpn_cls_loss = gluon.loss.SigmoidBinaryCrossEntropyLoss(from_sigmoid=False)
        rpn_box_loss = gluon.loss.HuberLoss(rho=self._cfg.train.rpn_smoothl1_rho)  # == smoothl1
        rcnn_cls_loss = gluon.loss.SoftmaxCrossEntropyLoss()
        rcnn_box_loss = gluon.loss.HuberLoss(rho=self._cfg.train.rcnn_smoothl1_rho)  # == smoothl1
        metrics = [mx.metric.Loss('RPN_Conf'),
                   mx.metric.Loss('RPN_SmoothL1'),
                   mx.metric.Loss('RCNN_CrossEntropy'),
                   mx.metric.Loss('RCNN_SmoothL1')]

        rpn_acc_metric = RPNAccMetric()
        rpn_bbox_metric = RPNL1LossMetric()
        rcnn_acc_metric = RCNNAccMetric()
        rcnn_bbox_metric = RCNNL1LossMetric()
        metrics2 = [rpn_acc_metric, rpn_bbox_metric, rcnn_acc_metric, rcnn_bbox_metric]

        # lr decay policy
        lr_decay = float(self._cfg.train.lr_decay)
        lr_steps = sorted(
            [float(ls) for ls in self._cfg.train.lr_decay_epoch])
        lr_warmup = float(self._cfg.train.lr_warmup)  # avoid int division

        if self._cfg.train.verbose:
            self._logger.info('Trainable parameters:')
            self._logger.info(self.net.collect_train_params().keys())
        self._logger.info('Start training from [Epoch %d]', max(self._cfg.train.start_epoch, self.epoch))
        self.net.collect_params().reset_ctx(self.ctx)
        self.net.target_generator.collect_params().reset_ctx(self.ctx)

        for self.epoch in range(max(self._cfg.train.start_epoch, self.epoch), self._cfg.train.epochs):
            epoch = self.epoch
            btic = time.time()
            rcnn_task = ForwardBackwardTask(self.net, self.trainer, rpn_cls_loss, rpn_box_loss,
                                            rcnn_cls_loss, rcnn_box_loss, mix_ratio=1.0,
                                            amp_enabled=self._cfg.faster_rcnn.amp)
            executor = Parallel(self._cfg.train.executor_threads,
                                rcnn_task) if not self._cfg.horovod else None
            mix_ratio = 1.0
            if not self._cfg.disable_hybridization:
                self.net.hybridize(static_alloc=self._cfg.faster_rcnn.static_alloc)
            if self._cfg.train.mixup:
                # TODO(zhreshold) only support evenly mixup now, target generator needs to be
                #  modified otherwise
                train_data._dataset._data.set_mixup(np.random.uniform, 0.5, 0.5)
                mix_ratio = 0.5
                if epoch >= self._cfg.train.epochs - self._cfg.train.no_mixup_epochs:
                    train_data._dataset._data.set_mixup(None)
                    mix_ratio = 1.0
            while lr_steps and epoch >= lr_steps[0]:
                new_lr = self.trainer.learning_rate * lr_decay
                lr_steps.pop(0)
                self.trainer.set_learning_rate(new_lr)
                self._logger.info("[Epoch %d] Set learning rate to %f", epoch, new_lr)
            for metric in metrics:
                metric.reset()
            tic = time.time()
            base_lr = self.trainer.learning_rate
            rcnn_task.mix_ratio = mix_ratio
            for i, batch in enumerate(train_data):
                if epoch == 0 and i <= lr_warmup:
                    # adjust based on real percentage
                    new_lr = base_lr * _get_lr_at_iter(i / lr_warmup,
                                                       self._cfg.train.lr_warmup_factor)
                    if new_lr != self.trainer.learning_rate:
                        if i % self._cfg.train.log_interval == 0:
                            self._logger.info(
                                '[Epoch 0 Iteration %d] Set learning rate to %f', i, new_lr)
                        self.trainer.set_learning_rate(new_lr)
                batch = _split_and_load(batch, ctx_list=self.ctx)
                metric_losses = [[] for _ in metrics]
                add_losses = [[] for _ in metrics2]
                if executor is not None:
                    for data in zip(*batch):
                        executor.put(data)
                for _ in range(len(self.ctx)):
                    if executor is not None:
                        result = executor.get()
                    else:
                        result = rcnn_task.forward_backward(list(zip(*batch))[0])
                    if (not self._cfg.horovod) or hvd.rank() == 0:
                        for k, metric_loss in enumerate(metric_losses):
                            metric_loss.append(result[k])
                        for k, add_loss in enumerate(add_losses):
                            add_loss.append(result[len(metric_losses) + k])
                for metric, record in zip(metrics, metric_losses):
                    metric.update(0, record)
                for metric, records in zip(metrics2, add_losses):
                    for pred in records:
                        metric.update(pred[0], pred[1])
                self.trainer.step(self.batch_size)

                # update metrics
                if (not self._cfg.horovod or hvd.rank() == 0) and self._cfg.train.log_interval \
                        and not (i + 1) % self._cfg.train.log_interval:
                    msg = ','.join(
                        ['{}={:.3f}'.format(*metric.get()) for metric in
                         metrics + metrics2])
                    self._logger.info('[Epoch {}][Batch {}], Speed: {:.3f} samples/sec, {}'.format(
                        epoch, i,
                        self._cfg.train.log_interval * self._cfg.train.batch_size / (
                            time.time() - btic), msg))
                    btic = time.time()

            if not self._cfg.horovod or hvd.rank() == 0:
                msg = ','.join(['{}={:.3f}'.format(*metric.get()) for metric in metrics])
                # pylint: disable=logging-format-interpolation
                self._logger.info('[Epoch {}] Training cost: {:.3f}, {}'.format(
                    epoch, (time.time() - tic), msg))
                if not (epoch + 1) % self._cfg.valid.val_interval:
                    # consider reduce the frequency of validation to save time
                    map_name, mean_ap = self._evaluate(val_data)
                    val_msg = '\n'.join(['{}={}'.format(k, v) for k, v in zip(map_name, mean_ap)])
                    self._logger.info('[Epoch {}] Validation: \n{}'.format(epoch, val_msg))
                    current_map = float(mean_ap[-1])
                    if current_map > self._best_map:
                        cp_name = os.path.join(self._logdir, 'best_checkpoint.pkl')
                        self._logger.info('[Epoch %d] Current best map: %f vs previous %f, saved to %s',
                                          self.epoch, current_map, self._best_map, cp_name)
                        self.save(cp_name)
                        self._best_map = current_map
                if self._reporter:
                    self._reporter(epoch=epoch, map_reward=current_map)
            self._time_elapsed += time.time() - btic
        # map on train data
        map_name, mean_ap = self._evaluate(train_eval_data)
        return {'train_map': float(mean_ap[-1]), 'valid_map': self._best_map, 'time': self._time_elapsed}

    def _evaluate(self, val_data):
        """Evaluate on validation dataset."""
        clipper = BBoxClipToImage()
        if not isinstance(val_data, gluon.data.DataLoader):
            if hasattr(val_data, 'to_mxnet'):
                val_data = val_data.to_mxnet()
            val_bfn = Tuple(*[Append() for _ in range(3)])
            short = self.net.short[-1] if isinstance(self.net.short, (tuple, list)) else self.net.short
            # validation use 1 sample per device
            val_data = gluon.data.DataLoader(
                val_data.transform(FasterRCNNDefaultValTransform(short, self.net.max_size)),
                len(self.ctx), False, batchify_fn=val_bfn, last_batch='keep',
                num_workers=self._cfg.num_workers)
        if self._cfg.valid.metric == 'voc07':
            eval_metric = VOC07MApMetric(iou_thresh=self._cfg.valid.iou_thresh, class_names=self.classes)
        elif self._cfg.valid.metric == 'voc':
            eval_metric = VOCMApMetric(iou_thresh=self._cfg.valid.iou_thresh, class_names=self.classes)
        else:
            raise ValueError(f'Invalid metric type: {self._cfg.valid.metric}')

        if not self._cfg.disable_hybridization:
            # input format is differnet than training, thus rehybridization is needed.
            self.net.hybridize(static_alloc=self._cfg.faster_rcnn.static_alloc)
        for batch in val_data:
            batch = _split_and_load(batch, ctx_list=self.ctx)
            det_bboxes = []
            det_ids = []
            det_scores = []
            gt_bboxes = []
            gt_ids = []
            gt_difficults = []
            for x, y, im_scale in zip(*batch):
                # get prediction results
                ids, scores, bboxes = self.net(x)
                det_ids.append(ids)
                det_scores.append(scores)
                # clip to image size
                det_bboxes.append(clipper(bboxes, x))
                # rescale to original resolution
                im_scale = im_scale.reshape((-1)).asscalar()
                det_bboxes[-1] *= im_scale
                # split ground truths
                gt_ids.append(y.slice_axis(axis=-1, begin=4, end=5))
                gt_bboxes.append(y.slice_axis(axis=-1, begin=0, end=4))
                gt_bboxes[-1] *= im_scale
                gt_difficults.append(
                    y.slice_axis(axis=-1, begin=5, end=6) if y.shape[-1] > 5 else None)
            # update metric
            for det_bbox, det_id, det_score, gt_bbox, gt_id, gt_diff in zip(det_bboxes, det_ids,
                                                                            det_scores, gt_bboxes,
                                                                            gt_ids, gt_difficults):
                eval_metric.update(det_bbox, det_id, det_score, gt_bbox, gt_id, gt_diff)
        return eval_metric.get()

    def _predict(self, x):
        """Predict an individual example."""
        short_size = self.net.short[-1] if isinstance(self.net.short, (tuple, list)) else self.net.short
        if isinstance(x, str):
            x = load_test(x, short=short_size, max_size=1024)[0]
        elif isinstance(x, mx.nd.NDArray):
            x = transform_test(x, short=short_size, max_size=1024)[0]
        elif isinstance(x, pd.DataFrame):
            assert 'image' in x.columns, "Expect column `image` for input images"
            def _predict_merge(x):
                y = self._predict(x)
                y['image'] = x
                return y
            return pd.concat([_predict_merge(xx) for xx in x['image']]).reset_index(drop=True)
        else:
            raise ValueError('Input is not supported: {}'.format(type(x)))
        height, width = x.shape[2:4]
        x = x.as_in_context(self.ctx[0])
        ids, scores, bboxes = [xx[0].asnumpy() for xx in self.net(x)]
        bboxes[:, (0, 2)] /= width
        bboxes[:, (1, 3)] /= height
        bboxes = np.clip(bboxes, 0.0, 1.0).tolist()
        df = pd.DataFrame({'predict_class': [self.classes[int(id)] for id in ids], 'predict_score': scores.flatten(),
                           'predict_rois': [{'xmin': bbox[0], 'ymin': bbox[1], 'xmax': bbox[2], 'ymax': bbox[3]} \
                                for bbox in bboxes]})
        # filter out invalid (scores < 0) rows
        valid_df = df[df['predict_score'] > 0].reset_index(drop=True)
        return valid_df

    def _init_network(self):
        if not self.num_class:
            raise ValueError('Unable to create network when `num_class` is unknown. \
                It should be inferred from dataset or resumed from saved states.')
        assert len(self.classes) == self.num_class

        # training contexts
        if self._cfg.horovod:
            self.ctx = [mx.gpu(hvd.local_rank())]
        else:
            ctx = [mx.gpu(int(i)) for i in self._cfg.gpus]
            self.ctx = ctx if ctx else [mx.cpu()]
        # network
        kwargs = {}
        module_list = []
        if self._cfg.faster_rcnn.use_fpn:
            module_list.append('fpn')
        if self._cfg.faster_rcnn.norm_layer is not None:
            module_list.append(self._cfg.faster_rcnn.norm_layer)
            if self._cfg.faster_rcnn.norm_layer == 'syncbn':
                kwargs['num_devices'] = len(self.ctx)

        self.num_gpus = hvd.size() if self._cfg.horovod else len(self.ctx)

        if self._cfg.faster_rcnn.transfer is not None:
            assert isinstance(self._cfg.faster_rcnn.transfer, str)
            self._logger.info(
                f'Using transfer learning from {self._cfg.faster_rcnn.transfer}, ' +
                'the other network parameters are ignored.')
            self._cfg.faster_rcnn.use_fpn = 'fpn' in self._cfg.faster_rcnn.transfer
            self.net = get_model(self._cfg.faster_rcnn.transfer, pretrained=True,
                                 per_device_batch_size=self._cfg.train.batch_size // self.num_gpus,
                                 **kwargs)
            self.net.reset_class(self.classes,
                                 reuse_weights=[cname for cname in self.classes if cname in self.net.classes])
        else:
            self._cfg.faster_rcnn.use_fpn = True
            if self._cfg.faster_rcnn.norm_layer == 'syncbn':
                norm_layer = gluon.contrib.nn.SyncBatchNorm
                norm_kwargs = {'num_devices': len(self.ctx)}
                sym_norm_layer = mx.sym.contrib.SyncBatchNorm
                sym_norm_kwargs = {'ndev': len(self.ctx)}
            elif self._cfg.faster_rcnn.norm_layer == 'gn':
                norm_layer = gluon.nn.GroupNorm
                norm_kwargs = {'groups': 8}
                sym_norm_layer = mx.sym.GroupNorm
                sym_norm_kwargs = {'groups': 8}
            else:
                norm_layer = gluon.nn.BatchNorm
                norm_kwargs = None
                sym_norm_layer = None
                sym_norm_kwargs = None
            self.net = get_model('custom_faster_rcnn_fpn', classes=self.classes, transfer=None,
                                 dataset=self._cfg.dataset,
                                 pretrained_base=self._cfg.train.pretrained_base,
                                 base_network_name=self._cfg.faster_rcnn.backbone,
                                 norm_layer=norm_layer, norm_kwargs=norm_kwargs,
                                 sym_norm_layer=sym_norm_layer, sym_norm_kwargs=sym_norm_kwargs,
                                 num_fpn_filters=self._cfg.faster_rcnn.num_fpn_filters,
                                 num_box_head_conv=self._cfg.faster_rcnn.num_box_head_conv,
                                 num_box_head_conv_filters=
                                 self._cfg.faster_rcnn.num_box_head_conv_filters,
                                 num_box_head_dense_filters=
                                 self._cfg.faster_rcnn.num_box_head_dense_filters,
                                 short=self._cfg.faster_rcnn.image_short,
                                 max_size=self._cfg.faster_rcnn.image_max_size,
                                 min_stage=2, max_stage=6,
                                 nms_thresh=self._cfg.faster_rcnn.nms_thresh,
                                 nms_topk=self._cfg.faster_rcnn.nms_topk,
                                 roi_mode=self._cfg.faster_rcnn.roi_mode,
                                 roi_size=self._cfg.faster_rcnn.roi_size,
                                 strides=self._cfg.faster_rcnn.strides,
                                 clip=self._cfg.faster_rcnn.clip,
                                 rpn_channel=self._cfg.faster_rcnn.rpn_channel,
                                 base_size=self._cfg.faster_rcnn.anchor_base_size,
                                 scales=self._cfg.faster_rcnn.anchor_scales,
                                 ratios=self._cfg.faster_rcnn.anchor_aspect_ratio,
                                 alloc_size=self._cfg.faster_rcnn.anchor_alloc_size,
                                 rpn_nms_thresh=self._cfg.faster_rcnn.rpn_nms_thresh,
                                 rpn_train_pre_nms=self._cfg.train.rpn_train_pre_nms,
                                 rpn_train_post_nms=self._cfg.train.rpn_train_post_nms,
                                 rpn_test_pre_nms=self._cfg.valid.rpn_test_pre_nms,
                                 rpn_test_post_nms=self._cfg.valid.rpn_test_post_nms,
                                 rpn_min_size=self._cfg.train.rpn_min_size,
                                 per_device_batch_size=self._cfg.train.batch_size // self.num_gpus,
                                 num_sample=self._cfg.train.rcnn_num_samples,
                                 pos_iou_thresh=self._cfg.train.rcnn_pos_iou_thresh,
                                 pos_ratio=self._cfg.train.rcnn_pos_ratio,
                                 max_num_gt=self._cfg.faster_rcnn.max_num_gt)

        if self._cfg.resume.strip():
            self.net.load_parameters(self._cfg.resume.strip())
        else:
            for param in self.net.collect_params().values():
                if param._data is not None:
                    continue
                param.initialize()
        self.net.collect_params().reset_ctx(self.ctx)
        if self._cfg.faster_rcnn.amp:
            # Cast both weights and gradients to 'float16'
            self.net.cast('float16')
            # These layers don't support type 'float16'
            self.net.collect_params('.*batchnorm.*').setattr('dtype', 'float32')
            self.net.collect_params('.*normalizedperclassboxcenterencoder.*').setattr('dtype', 'float32')
        if self._cfg.resume.strip():
            self.net.load_parameters(self._cfg.resume.strip())
        else:
            for param in self.net.collect_params().values():
                if param._data is not None:
                    continue
                param.initialize()
        self.net.collect_params().reset_ctx(self.ctx)

    def _init_trainer(self):
        kv_store_type = 'device' if (self._cfg.faster_rcnn.amp and 'nccl' in self._cfg.kv_store) \
            else self._cfg.kv_store
        kv = mx.kvstore.create(kv_store_type)
        optimizer_params = {'learning_rate': self._cfg.train.lr, 'wd': self._cfg.train.wd,
                            'momentum': self._cfg.train.momentum}
        if self._cfg.faster_rcnn.amp:
            optimizer_params['multi_precision'] = True
        if self._cfg.horovod:
            hvd.broadcast_parameters(self.net.collect_params(), root_rank=0)
            self.trainer = hvd.DistributedTrainer(
                self.net.collect_train_params(),  # fix batchnorm, fix first stage, etc...
                'sgd',
                optimizer_params)
        else:
            self.trainer = gluon.Trainer(
                self.net.collect_train_params(),  # fix batchnorm, fix first stage, etc...
                'sgd',
                optimizer_params,
                update_on_kvstore=(False if self._cfg.faster_rcnn.amp else None), kvstore=kv)

        if self._cfg.faster_rcnn.amp:
            self._cfg.init_trainer(self.trainer)



# pylint: disable=line-too-long,too-many-lines,missing-docstring
"""HMDB51 action classification dataset.
http://serre-lab.clps.brown.edu/resource/hmdb-a-large-human-motion-database/.
Code adapted from https://github.com/open-mmlab/mmaction and
https://github.com/bryanyzhu/two-stream-pytorch"""
import os
from ..video_custom import VideoClsCustom

__all__ = ['HMDB51']

class HMDB51(VideoClsCustom):
    """Load the HMDB51 video action recognition dataset.

    Refer to :doc:`../build/examples_datasets/hmdb51` for the description of
    this dataset and how to prepare it.

    Parameters
    ----------
    root : str, required. Default '~/.mxnet/datasets/hmdb51'
        Path to the root folder storing the dataset.
    setting : str, required.
        A text file describing the dataset, each line per video sample.
        There are three items in each line: (1) video path; (2) video length and (3) video label.
    train : bool, default True.
        Whether to load the training or validation set.
    test_mode : bool, default False.
        Whether to perform evaluation on the test set.
        Usually there is three-crop or ten-crop evaluation strategy involved.
    name_pattern : str, default None.
        The naming pattern of the decoded video frames.
        For example, img_00012.jpg.
    video_ext : str, default 'mp4'.
        If video_loader is set to True, please specify the video format accordinly.
    is_color : bool, default True.
        Whether the loaded image is color or grayscale.
    modality : str, default 'rgb'.
        Input modalities, we support only rgb video frames for now.
        Will add support for rgb difference image and optical flow image later.
    num_segments : int, default 1.
        Number of segments to evenly divide the video into clips.
        A useful technique to obtain global video-level information.
        Limin Wang, etal, Temporal Segment Networks: Towards Good Practices for Deep Action Recognition, ECCV 2016.
    num_crop : int, default 1.
        Number of crops for each image. default is 1.
        Common choices are three crops and ten crops during evaluation.
    new_length : int, default 1.
        The length of input video clip. Default is a single image, but it can be multiple video frames.
        For example, new_length=16 means we will extract a video clip of consecutive 16 frames.
    new_step : int, default 1.
        Temporal sampling rate. For example, new_step=1 means we will extract a video clip of consecutive frames.
        new_step=2 means we will extract a video clip of every other frame.
    new_width : int, default 340.
        Scale the width of loaded image to 'new_width' for later multiscale cropping and resizing.
    new_height : int, default 256.
        Scale the height of loaded image to 'new_height' for later multiscale cropping and resizing.
    target_width : int, default 224.
        Scale the width of transformed image to the same 'target_width' for batch forwarding.
    target_height : int, default 224.
        Scale the height of transformed image to the same 'target_height' for batch forwarding.
    temporal_jitter : bool, default False.
        Whether to temporally jitter if new_step > 1.
    video_loader : bool, default False.
        Whether to use video loader to load data.
    use_decord : bool, default True.
        Whether to use Decord video loader to load data. Otherwise use mmcv video loader.
    transform : function, default None.
        A function that takes data and label and transforms them.
    slowfast : bool, default False.
        If set to True, use data loader designed for SlowFast network.
        Christoph Feichtenhofer, etal, SlowFast Networks for Video Recognition, ICCV 2019.
    slow_temporal_stride : int, default 16.
        The temporal stride for sparse sampling of video frames in slow branch of a SlowFast network.
    fast_temporal_stride : int, default 2.
        The temporal stride for sparse sampling of video frames in fast branch of a SlowFast network.
    data_aug : str, default 'v1'.
        Different types of data augmentation auto. Supports v1, v2, v3 and v4.
    lazy_init : bool, default False.
        If set to True, build a dataset instance without loading any dataset.
    """
    def __init__(self,
                 root=os.path.expanduser('~/.mxnet/datasets/hmdb51/rawframes'),
                 setting=os.path.expanduser('~/.mxnet/datasets/hmdb51/testTrainMulti_7030_splits/hmdb51_train_split_1_rawframes.txt'),
                 train=True,
                 test_mode=False,
                 name_pattern='img_%05d.jpg',
                 video_ext='mp4',
                 is_color=True,
                 modality='rgb',
                 num_segments=1,
                 num_crop=1,
                 new_length=1,
                 new_step=1,
                 new_width=340,
                 new_height=256,
                 target_width=224,
                 target_height=224,
                 temporal_jitter=False,
                 video_loader=False,
                 use_decord=False,
                 slowfast=False,
                 slow_temporal_stride=16,
                 fast_temporal_stride=2,
                 data_aug='v1',
                 lazy_init=False,
                 transform=None):

        super(HMDB51, self).__init__(root, setting, train, test_mode, name_pattern,
                                     video_ext, is_color, modality, num_segments,
                                     num_crop, new_length, new_step, new_width, new_height,
                                     target_width, target_height, temporal_jitter,
                                     video_loader, use_decord, slowfast, slow_temporal_stride,
                                     fast_temporal_stride, data_aug, lazy_init, transform)

class HMDB51Attr(object):
    def __init__(self):
        self.num_class = 51
        self.classes = ['brush_hair', 'cartwheel', 'catch', 'chew', 'clap', 'climb', 'climb_stairs',
                        'dive', 'draw_sword', 'dribble', 'drink', 'eat', 'fall_floor', 'fencing',
                        'flic_flac', 'golf', 'handstand', 'hit', 'hug', 'jump', 'kick', 'kick_ball',
                        'kiss', 'laugh', 'pick', 'pour', 'pullup', 'punch', 'push', 'pushup',
                        'ride_bike', 'ride_horse', 'run', 'shake_hands', 'shoot_ball', 'shoot_bow',
                        'shoot_gun', 'sit', 'situp', 'smile', 'smoke', 'somersault', 'stand',
                        'swing_baseball', 'sword', 'sword_exercise', 'talk', 'throw', 'turn', 'walk', 'wave']



import numbers
import cv2
import numpy as np
import PIL

import torch


def _is_tensor_clip(clip):
    return torch.is_tensor(clip) and clip.ndimension() == 4


def crop_clip(clip, min_h, min_w, h, w):
    if isinstance(clip[0], np.ndarray):
        cropped = [img[min_h:min_h + h, min_w:min_w + w, :] for img in clip]

    elif isinstance(clip[0], PIL.Image.Image):
        cropped = [
            img.crop((min_w, min_h, min_w + w, min_h + h)) for img in clip
        ]
    else:
        raise TypeError('Expected numpy.ndarray or PIL.Image' +
                        'but got list of {0}'.format(type(clip[0])))
    return cropped


def resize_clip(clip, size, interpolation='bilinear'):
    if isinstance(clip[0], np.ndarray):
        if isinstance(size, numbers.Number):
            im_h, im_w, im_c = clip[0].shape
            # Min spatial dim already matches minimal size
            if (im_w <= im_h and im_w == size) or (im_h <= im_w
                                                   and im_h == size):
                return clip
            new_h, new_w = get_resize_sizes(im_h, im_w, size)
            size = (new_w, new_h)
        else:
            size = size[0], size[1]
        if interpolation == 'bilinear':
            np_inter = cv2.INTER_LINEAR
        else:
            np_inter = cv2.INTER_NEAREST
        scaled = [
            cv2.resize(img, size, interpolation=np_inter) for img in clip
        ]
    elif isinstance(clip[0], PIL.Image.Image):
        if isinstance(size, numbers.Number):
            im_w, im_h = clip[0].size
            # Min spatial dim already matches minimal size
            if (im_w <= im_h and im_w == size) or (im_h <= im_w
                                                   and im_h == size):
                return clip
            new_h, new_w = get_resize_sizes(im_h, im_w, size)
            size = (new_w, new_h)
        else:
            size = size[1], size[0]
        if interpolation == 'bilinear':
            pil_inter = PIL.Image.BILINEAR
        else:
            pil_inter = PIL.Image.NEAREST
        scaled = [img.resize(size, pil_inter) for img in clip]
    else:
        raise TypeError('Expected numpy.ndarray or PIL.Image' +
                        'but got list of {0}'.format(type(clip[0])))
    return scaled


def get_resize_sizes(im_h, im_w, size):
    if im_w < im_h:
        ow = size
        oh = int(size * im_h / im_w)
    else:
        oh = size
        ow = int(size * im_w / im_h)
    return oh, ow


def normalize(clip, mean, std, inplace=False):
    if not _is_tensor_clip(clip):
        raise TypeError('tensor is not a torch clip.')

    if not inplace:
        clip = clip.clone()

    dtype = clip.dtype
    mean = torch.as_tensor(mean, dtype=dtype, device=clip.device)
    std = torch.as_tensor(std, dtype=dtype, device=clip.device)
    clip.sub_(mean[:, None, None, None]).div_(std[:, None, None, None])

    return clip



import glob
import numpy as np
import os
import re
import tempfile
from addict import Dict
from collections import defaultdict
from pathlib import Path
from subprocess import call
import cv2

from ..dataset import GluonCVMotionDataset, DataSample, FieldNames, get_resized_video_location
from ..io.video_io import VideoFrameReader, VideoSortedFolderReader, cmd_with_addn_args, \
    FFMPEG_FOURCC


def default_filepath_filter(path):
    return not path.name.startswith('.')


def get_filepaths(root_path, ext=".mp4", replace_path=None, filter_fn=default_filepath_filter,
                  glob_exp=None, parents_only=False):
    if replace_path is None:
        replace_path = root_path
    if glob_exp is None:
        glob_exp = '*{}'.format(ext)
    paths = glob.glob(str(Path(root_path) / "**" / glob_exp), recursive=True)
    paths = [Path(p) for p in paths]
    if replace_path:
        rel_paths = [p.relative_to(replace_path) for p in paths]
    else:
        rel_paths = paths
    if filter_fn:
        rel_paths = [p for p in rel_paths if filter_fn(p)]
    if parents_only:
        parents = {rel_path.parent for rel_path in rel_paths}
        rel_paths = list(parents)
    return rel_paths


def process_dataset_splits(dataset, sample_split_fn, save=False, log=True):
    splits = defaultdict(list)
    for sample_id, sample in dataset.samples:
        sample_split = sample_split_fn(sample)
        if sample_split is not None:
            splits[sample_split].append(sample_id)

    if log:
        print("Processed {} samples".format(len(dataset.samples)))
        for split_name, split_ids in splits.items():
            print("Found {} samples for {}".format(len(split_ids), split_name))

    if save:
        if log:
            print("Saving data splits to: {}".format(dataset.split_path))
        dataset.splits = splits
        dataset.dump_splits()

    return splits


def get_datasample_from_video(rel_path, full_path=None, dataset=None):
    if full_path is None:
        if dataset is not None:
            full_path = os.path.join(dataset.data_root_path, rel_path)
        else:
            raise ValueError("Both full path and dataset cannot be empty")
    vid_reader = VideoFrameReader(full_path)

    sample = DataSample(rel_path)

    fps = vid_reader.fps
    num_frames = len(vid_reader)
    duration = vid_reader.duration
    frame = vid_reader.read()
    if frame is None:
        return None
    width = frame.width
    height = frame.height

    metadata = {
        FieldNames.DATA_PATH: rel_path,
        FieldNames.FPS: fps,
        FieldNames.DURATION: duration,
        FieldNames.NUM_FRAMES: num_frames,
        FieldNames.RESOLUTION: {"width": width, "height": height},
    }
    sample.metadata = metadata

    return sample


def check_dimensions(dataset, target_sidelen=256, cache_func=None):
    if cache_func is None:
        cache_func = lambda sample: get_resized_video_location(sample, target_sidelen)
    bad_vids = []
    for i, (id, sample) in enumerate(dataset.samples):
        vid_path = cache_func(sample)
        if not os.path.exists(vid_path):
            print("Warning, missing cache vid path {} , skipping".format(vid_path))
            continue

        vid_reader = VideoFrameReader(vid_path)
        if target_sidelen not in (vid_reader.width, vid_reader.height):
            print("Warning, target side length {} not in video, actual res: {}x{} , \
                vid path: {}".format(
                target_sidelen, vid_reader.width, vid_reader.height, vid_path))
            bad_vids.append(id)

        if (i + 1) % 100 == 0:
            print("Done checking {} vids".format(i + 1))

    print("Total of {} bad sized vids".format(len(bad_vids)))
    print(bad_vids)
    return bad_vids


def crop_fn_ffmpeg(input_path, output_path, t1, t2, fps=None, addn_args=None):
    t1_secs, t2_secs = [t / 1000.0 for t in (t1, t2)]

    # You need -strict experimental on some versions of ffmpeg (including the ubuntu default) 
    # or it will throw an error
    # when decoding aac audio, see: https://stackoverflow.com/a/32932092 
    # and https://superuser.com/a/543593
    cmd = [
        "ffmpeg",
        "-y",
        "-hide_banner",
        "-loglevel", "error",
        "-ss", "{:.3f}".format(t1_secs),
        "-i", input_path,
        "-t", "{:.3f}".format(t2_secs - t1_secs),
    ]
    if fps is not None:
        cmd += ["-r", str(fps)]
    cmd = cmd_with_addn_args(cmd, addn_args)
    cmd += ["-strict", "experimental", output_path]

    return call(cmd)


def crop_fn_videoframereader_opencv(input_path, output_path, t1, t2):
    start_time, end_time = t1, t2

    frame_reader = VideoFrameReader(input_path)
    fps = frame_reader.fps
    ts = start_time
    frame = frame_reader.get_frame_for_time(ts)

    # We are using the VideoWriter directly instead of write_video_rgb 
    # due to memory issues of holding all frames
    # in memory before writing
    writer = cv2.VideoWriter(output_path, FFMPEG_FOURCC['avc1'], fps=fps,
                             frameSize=(frame.width, frame.height), isColor=True)
    while ts <= end_time and frame is not None:
        frame = np.array(frame)
        frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
        writer.write(frame)

        ts = frame_reader.get_next_frame_time_stamp()
        frame = frame_reader.read()

    writer.release()

    return True


def crop_fn_frame_folder(input_path, output_path, t1, t2, in_fps=None,
                         frame_reader=None, symlink=True,
                         copy_names=True, copy_frames=True):
    if in_fps is None and frame_reader is None:
        raise ValueError("Both in_fps and frame_reader can't be None")

    if frame_reader is None:
        frame_reader = VideoSortedFolderReader(input_path, fps=in_fps)

    start_time, end_time = t1, t2

    ts = start_time
    image_ext = frame_reader.get_image_ext()

    output_path = Path(output_path)
    output_path.mkdir(exist_ok=True)
    for frame_idx, (frame, ts) in enumerate(frame_reader.iter_frame_ts(ts)):
        if ts > end_time:
            break

        frame_path = frame_reader.get_frame_path()
        if copy_names:
            im_filename = Path(frame_path).name
        else:
            # 6 digits is about 9 hours of video at 30 fps
            im_filename = f"{frame_idx:06}" + image_ext
        im_out_path = Path(output_path, im_filename)
        if symlink:
            os.symlink(frame_path, im_out_path)
        elif copy_frames:
            call(["cp", frame_path, im_out_path])
        else:
            frame.save(im_out_path)

    return True


def crop_video(input_path, output_path, start_time, end_time,
               overwrite=False, min_video_length_ms=1000,
               crop_fn=crop_fn_ffmpeg, add_output_ts=False,
               make_dirs=False, *args, **kwargs):
    if add_output_ts:
        output_base, ext = os.path.splitext(output_path)
        output_path = "{name}_{t1}-{t2}{ext}".format(name=output_base,
                                                     t1=start_time,
                                                     t2=end_time,
                                                     ext=ext)

    if not os.path.exists(input_path):
        raise ValueError("input video path: {} does not exist".format(input_path))

    if end_time - start_time < min_video_length_ms:
        raise ValueError("End time - start time must be >= {min_len}, \
                          error for video: {path}".format(
                            min_len=min_video_length_ms, path=input_path))

    parent_dir = os.path.dirname(output_path)
    if not os.path.exists(parent_dir):
        if make_dirs:
            os.makedirs(parent_dir, exist_ok=True)
        else:
            raise ValueError("Parent output directory for {} \
                              does not exist for input video path: {}" .format(
                                output_path, input_path))

    if os.path.exists(output_path) and (not overwrite):
        return False, output_path

    with tempfile.TemporaryDirectory() as tmp_crop_dir_path:
        tmp_crop_path = os.path.join(tmp_crop_dir_path, os.path.basename(output_path))
        crop_fn(input_path, tmp_crop_path, start_time, end_time, *args, **kwargs)

        ret_code = call(['mv', tmp_crop_path, output_path])
        if ret_code:
            raise Exception('Failed to move to {} from tmp location'.format(output_path))

    return True, output_path


class AddictList(Dict):
    """Class that can be used for gathering statistics 
    about data by easily storing lists of different data"""

    LKEY = "lst"

    def _get_lst(self):
        lkey = self.LKEY
        if lkey not in self:
            self[lkey] = []
        return self[lkey]

    def append(self, item):
        self._get_lst().append(item)

    def extend(self, lst2):
        self._get_lst().extend(lst2)

    def replace_lst(self):
        for k, v in self.items():
            if isinstance(v, type(self)):
                if len(v) == 1 and self.LKEY in v:
                    self[k] = v[self.LKEY]
                else:
                    v.replace_lst()


chunk_pat = re.compile(r"_\d+-\d+$")
def default_chunked_sample_to_orig_id(chunked_sample, chunk_subdir="", orig_dataset=None):
    id_path = Path(chunked_sample.id)
    chunked_id_stem = id_path.stem
    orig_id = chunk_pat.sub('', chunked_id_stem, count=1)
    orig_id = str(id_path.with_name(orig_id + id_path.suffix).relative_to(chunk_subdir))
    return orig_id


def get_orig_chunked_sample(sample: DataSample, orig_dataset: GluonCVMotionDataset,
                            chunk_subdirs: [] = None,
                            sample_map_func=default_chunked_sample_to_orig_id,
                            exclude_unchunked=False):
    if chunk_subdirs:
        chunk_subdir = [subdir for subdir in chunk_subdirs \
                        if sample.data_relative_path.startswith(str(subdir))]
        if not chunk_subdir:
            return None
        chunk_subdir = str(chunk_subdir[0])
    else:
        chunk_subdir = Path(sample.data_relative_path).parts[0]
    orig_id = sample_map_func(sample, chunk_subdir, orig_dataset)
    if orig_id not in orig_dataset:
        return None
    orig_sample = orig_dataset[orig_id]
    if exclude_unchunked and \
            Path(orig_sample.data_relative_path) == \
            Path(sample.data_relative_path).relative_to(chunk_subdir):
        return None
    return orig_sample


def get_chunked_id_map(chunked_dataset, orig_dataset, *args, **kwargs):
    id_map = {}
    for chunked_id, sample in chunked_dataset.samples:
        orig_sample = get_orig_chunked_sample(sample, orig_dataset, *args, **kwargs)
        orig_id = orig_sample.id
        if orig_id is not None:
            id_map[chunked_id] = orig_id
    return id_map



from tqdm import tqdm
import unittest
import numpy as np
import mxnet as mx
from mxnet.test_utils import assert_almost_equal
from mxnet.gluon.data.vision import transforms

import gluoncv
from gluoncv.utils.metrics.segmentation import *
from gluoncv.data import ADE20KSegmentation

from .common import try_gpu, with_cpu

@unittest.skip("Temp disabled...")
@try_gpu(0)
def test_segmentation_utils():
    ctx = mx.context.current_context()
    import os
    if not os.path.isdir(os.path.expanduser('~/.mxnet/datasets/voc')):
        return

    transform_fn = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize([.485, .456, .406], [.229, .224, .225])
    ])
    # get the dataset
    # TODO FIXME: change it to ADE20K dataset and pretrained model
    dataset = ADE20KSegmentation(split='val')
    # load pretrained net
    net = gluoncv.model_zoo.get_model('fcn_resnet50_ade', pretrained=True, ctx=ctx)
    # count for pixAcc and mIoU
    total_inter, total_union, total_correct, total_label = 0, 0, 0, 0
    np_inter, np_union, np_correct, np_label = 0, 0, 0, 0
    tbar = tqdm(range(10))
    for i in tbar:
        img, mask = dataset[i]
        # prepare data and make prediction
        img = transform_fn(img)
        img = img.expand_dims(0).as_in_context(ctx)
        mask = mask.expand_dims(0)
        pred = net.evaluate(img).as_in_context(mx.cpu(0))
        # gcv prediction
        correct1, labeled1 = batch_pix_accuracy(pred, mask)
        inter1, union1 = batch_intersection_union(pred, mask, dataset.num_class)
        total_correct += correct1
        total_label += labeled1
        total_inter += inter1
        total_union += union1
        pixAcc = 1.0 * total_correct / (np.spacing(1) + total_label)
        IoU = 1.0 * total_inter / (np.spacing(1) + total_union)
        mIoU = IoU.mean()

        # np predicition
        pred2 = np.argmax(pred.asnumpy().astype('int64'), 1) + 1
        mask2 = mask.squeeze().asnumpy().astype('int64') + 1
        _, correct2, labeled2 = pixelAccuracy(pred2, mask2)
        inter2, union2 = intersectionAndUnion(pred2, mask2, dataset.num_class)
        np_correct += correct2
        np_label += labeled2
        np_inter += inter2
        np_union += union2
        np_pixAcc = 1.0 * np_correct / (np.spacing(1) + np_label)
        np_IoU = 1.0 * np_inter / (np.spacing(1) + np_union)
        np_mIoU = np_IoU.mean()
        tbar.set_description('pixAcc: %.3f, np_pixAcc: %.3f, mIoU: %.3f, np_mIoU: %.3f'%\
            (pixAcc, np_pixAcc, mIoU, np_mIoU))

    np.testing.assert_allclose(total_inter, np_inter)
    np.testing.assert_allclose(total_union, np_union)
    np.testing.assert_allclose(total_correct, np_correct)
    np.testing.assert_allclose(total_label, np_label)

if __name__ == '__main__':
    import nose
    nose.runmodule()



import numpy as np

cimport numpy as np
cimport cython

@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
def sampler(np.ndarray[float, ndim=2] ious, float pos_iou_thresh,float neg_iou_thresh,
            int num_sample, int max_pos, float eps):
    cdef int num_anchors = ious.shape[0]
    cdef int num_gt = ious.shape[1]
    cdef int num_pos = 0
    cdef int i, j, num_neg, ind
    cdef float ious_max_per_anchor
    cdef np.ndarray[float, ndim=1] samples = np.zeros((num_anchors,), dtype=np.float32)
    cdef np.ndarray[int, ndim=1] matches = np.zeros((num_anchors,), dtype=np.int32)
    cdef np.ndarray[float, ndim=1] ious_max_per_gt = np.max(ious, axis=0)
    cdef np.ndarray[int, ndim=1] indexes = np.random.permutation(num_anchors).astype(np.int32)

    with nogil:
        for i in range(num_anchors):
            ious_max_per_anchor = -1.0
            for j in range(num_gt):
                if ious_max_per_anchor < ious[i, j]:
                    ious_max_per_anchor = ious[i, j]  # max
                    matches[i] = j  # argmax
            for j in range(num_gt):
                if ious[i, j] + eps > ious_max_per_gt[j]:
                    samples[i] = 1.0
                    num_pos += 1
                    break
            if ious_max_per_anchor >= pos_iou_thresh and samples[i] < 1.0:
                samples[i] = 1.0
                num_pos += 1
            if 0.0 <= ious_max_per_anchor < neg_iou_thresh:
                if samples[i] + eps > 1.0:
                    num_pos -=1
                samples[i] = -1.0
        num_pos = min(num_pos, max_pos)
        num_neg = num_sample - num_pos
        for i in range(num_anchors):
            ind = indexes[i]
            if samples[ind] > 0.0 and num_pos > 0:
                num_pos -= 1
            elif samples[ind] < 0.0 and num_neg > 0:
                num_neg -= 1
            elif samples[ind] > 0.0 and num_pos <= 0:
                samples[ind] = 0.0
            elif samples[ind] < 0.0 and num_neg <= 0:
                samples[ind] = 0.0
    return samples, matches


import random

from .utils import functional as F


class Normalize(object):
    """Normalize a tensor image with mean and standard deviation

    Given mean: m and std: s
    will  normalize each channel as channel = (channel - mean) / std

    Args:
        mean (int): mean value
        std (int): std value
    """

    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, tensor):
        """
        Args:
            tensor (Tensor): Tensor of stacked images or image
            of size (C, H, W) to be normalized

        Returns:
            Tensor: Normalized stack of image of image
        """
        return F.normalize(tensor, self.mean, self.std)


class SpatialRandomCrop(object):
    """Crops a random spatial crop in a spatio-temporal
    numpy or tensor input [Channel, Time, Height, Width]
    """

    def __init__(self, size):
        """
        Args:
            size (tuple): in format (height, width)
        """
        self.size = size

    def __call__(self, tensor):
        h, w = self.size
        _, _, tensor_h, tensor_w = tensor.shape

        if w > tensor_w or h > tensor_h:
            error_msg = (
                'Initial tensor spatial size should be larger then '
                'cropped size but got cropped sizes : ({w}, {h}) while '
                'initial tensor is ({t_w}, {t_h})'.format(
                    t_w=tensor_w, t_h=tensor_h, w=w, h=h))
            raise ValueError(error_msg)
        x1 = random.randint(0, tensor_w - w)
        y1 = random.randint(0, tensor_h - h)
        cropped = tensor[:, :, y1:y1 + h, x1:x1 + h]
        return cropped



# pylint: disable=missing-function-docstring, line-too-long, unused-argument
"""
SlowFast Networks for Video Recognition
ICCV 2019, https://arxiv.org/abs/1812.03982
Code adapted from https://github.com/r1ch88/SlowFastNetworks
"""
import torch
import torch.nn as nn
from torch.nn import BatchNorm3d


__all__ = ['SlowFast', 'slowfast_4x16_resnet50_kinetics400', 'slowfast_8x8_resnet50_kinetics400',
           'slowfast_4x16_resnet101_kinetics400', 'slowfast_8x8_resnet101_kinetics400',
           'slowfast_16x8_resnet101_kinetics400', 'slowfast_16x8_resnet101_50_50_kinetics400',
           'slowfast_16x8_resnet50_sthsthv2', 'slowfast_4x16_resnet50_custom']


class Bottleneck(nn.Module):
    """Bottleneck block for ResNet-SlowFast.
    """
    expansion = 4

    def __init__(self,
                 inplanes,
                 planes,
                 strides=1,
                 downsample=None,
                 head_conv=1,
                 norm_layer=BatchNorm3d,
                 norm_kwargs=None,
                 layer_name=''):
        super(Bottleneck, self).__init__()

        if head_conv == 1:
            self.conv1 = nn.Conv3d(in_channels=inplanes, out_channels=planes, kernel_size=1, bias=False)
            self.bn1 = norm_layer(num_features=planes, **({} if norm_kwargs is None else norm_kwargs))
        elif head_conv == 3:
            self.conv1 = nn.Conv3d(in_channels=inplanes, out_channels=planes, kernel_size=(3, 1, 1), padding=(1, 0, 0),
                                   bias=False)
            self.bn1 = norm_layer(num_features=planes, **({} if norm_kwargs is None else norm_kwargs))
        else:
            raise ValueError("Unsupported head_conv!")
        self.conv2 = nn.Conv3d(in_channels=planes, out_channels=planes, kernel_size=(1, 3, 3),
                               stride=(1, strides, strides), padding=(0, 1, 1), bias=False)
        self.bn2 = norm_layer(num_features=planes, **({} if norm_kwargs is None else norm_kwargs))
        self.conv3 = nn.Conv3d(in_channels=planes, out_channels=planes * self.expansion, kernel_size=1, stride=1,
                               bias=False)
        self.bn3 = norm_layer(num_features=planes * self.expansion, **({} if norm_kwargs is None else norm_kwargs))
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample

    def forward(self, x):
        identity = x
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)
        out = self.conv3(out)
        out = self.bn3(out)
        if self.downsample is not None:
            identity = self.downsample(x)
        out = self.relu(out + identity)
        return out


class SlowFast(nn.Module):
    """SlowFast networks (SlowFast) from
    `"SlowFast Networks for Video Recognition"
    <https://arxiv.org/abs/1812.03982>`_ paper.
    """

    def __init__(self,
                 num_classes,
                 block=Bottleneck,
                 layers=None,
                 num_block_temp_kernel_fast=None,
                 num_block_temp_kernel_slow=None,
                 pretrained=False,
                 pretrained_base=False,
                 feat_ext=False,
                 num_segment=1,
                 num_crop=1,
                 bn_eval=True,
                 bn_frozen=False,
                 partial_bn=False,
                 frozen_stages=-1,
                 dropout_ratio=0.5,
                 init_std=0.01,
                 alpha=8,
                 beta_inv=8,
                 fusion_conv_channel_ratio=2,
                 fusion_kernel_size=5,
                 width_per_group=64,
                 num_groups=1,
                 slow_temporal_stride=16,
                 fast_temporal_stride=2,
                 slow_frames=4,
                 fast_frames=32,
                 norm_layer=BatchNorm3d,
                 norm_kwargs=None,
                 ctx=None,
                 **kwargs):
        super(SlowFast, self).__init__()
        self.num_segment = num_segment
        self.num_crop = num_crop
        self.dropout_ratio = dropout_ratio
        self.init_std = init_std
        self.alpha = alpha
        self.beta_inv = beta_inv
        self.fusion_conv_channel_ratio = fusion_conv_channel_ratio
        self.fusion_kernel_size = fusion_kernel_size
        self.width_per_group = width_per_group
        self.num_groups = num_groups
        self.dim_inner = self.num_groups * self.width_per_group
        self.out_dim_ratio = self.beta_inv // self.fusion_conv_channel_ratio
        self.slow_temporal_stride = slow_temporal_stride
        self.fast_temporal_stride = fast_temporal_stride
        self.slow_frames = slow_frames
        self.fast_frames = fast_frames
        self.feat_ext = feat_ext

        # build fast pathway
        self.fast_conv1 = nn.Conv3d(in_channels=3, out_channels=self.width_per_group // self.beta_inv,
                                    kernel_size=(5, 7, 7), stride=(1, 2, 2), padding=(2, 3, 3), bias=False)
        self.fast_bn1 = norm_layer(num_features=self.width_per_group // self.beta_inv,
                                   **({} if norm_kwargs is None else norm_kwargs))
        self.fast_relu = nn.ReLU(inplace=True)
        self.fast_maxpool = nn.MaxPool3d(kernel_size=(1, 3, 3), stride=(1, 2, 2), padding=(0, 1, 1))
        self.fast_res2 = self._make_layer_fast(inplanes=self.width_per_group // self.beta_inv,
                                               planes=self.dim_inner // self.beta_inv,
                                               num_blocks=layers[0],
                                               head_conv=3,
                                               norm_layer=norm_layer,
                                               norm_kwargs=norm_kwargs,
                                               layer_name='fast_res2_')
        self.fast_res3 = self._make_layer_fast(inplanes=self.width_per_group * 4 // self.beta_inv,
                                               planes=self.dim_inner * 2 // self.beta_inv,
                                               num_blocks=layers[1],
                                               strides=2,
                                               head_conv=3,
                                               norm_layer=norm_layer,
                                               norm_kwargs=norm_kwargs,
                                               layer_name='fast_res3_')
        self.fast_res4 = self._make_layer_fast(inplanes=self.width_per_group * 8 // self.beta_inv,
                                               planes=self.dim_inner * 4 // self.beta_inv,
                                               num_blocks=layers[2],
                                               num_block_temp_kernel_fast=num_block_temp_kernel_fast,
                                               strides=2,
                                               head_conv=3,
                                               norm_layer=norm_layer,
                                               norm_kwargs=norm_kwargs,
                                               layer_name='fast_res4_')
        self.fast_res5 = self._make_layer_fast(inplanes=self.width_per_group * 16 // self.beta_inv,
                                               planes=self.dim_inner * 8 // self.beta_inv,
                                               num_blocks=layers[3],
                                               strides=2,
                                               head_conv=3,
                                               norm_layer=norm_layer,
                                               norm_kwargs=norm_kwargs,
                                               layer_name='fast_res5_')

        # build lateral connections
        self.lateral_p1 = nn.Sequential(
            nn.Conv3d(in_channels=self.width_per_group // self.beta_inv,
                      out_channels=self.width_per_group // self.beta_inv * self.fusion_conv_channel_ratio,
                      kernel_size=(self.fusion_kernel_size, 1, 1),
                      stride=(self.alpha, 1, 1),
                      padding=(self.fusion_kernel_size // 2, 0, 0),
                      bias=False),
            norm_layer(num_features=self.width_per_group // self.beta_inv * self.fusion_conv_channel_ratio,
                       **({} if norm_kwargs is None else norm_kwargs)),
            nn.ReLU(inplace=True)
        )

        self.lateral_res2 = nn.Sequential(
            nn.Conv3d(in_channels=self.width_per_group * 4 // self.beta_inv,
                      out_channels=self.width_per_group * 4 // self.beta_inv * self.fusion_conv_channel_ratio,
                      kernel_size=(self.fusion_kernel_size, 1, 1),
                      stride=(self.alpha, 1, 1),
                      padding=(self.fusion_kernel_size // 2, 0, 0),
                      bias=False),
            norm_layer(num_features=self.width_per_group * 4 // self.beta_inv * self.fusion_conv_channel_ratio,
                       **({} if norm_kwargs is None else norm_kwargs)),
            nn.ReLU(inplace=True)
        )

        self.lateral_res3 = nn.Sequential(
            nn.Conv3d(in_channels=self.width_per_group * 8 // self.beta_inv,
                      out_channels=self.width_per_group * 8 // self.beta_inv * self.fusion_conv_channel_ratio,
                      kernel_size=(self.fusion_kernel_size, 1, 1),
                      stride=(self.alpha, 1, 1),
                      padding=(self.fusion_kernel_size // 2, 0, 0),
                      bias=False),
            norm_layer(num_features=self.width_per_group * 8 // self.beta_inv * self.fusion_conv_channel_ratio,
                       **({} if norm_kwargs is None else norm_kwargs)),
            nn.ReLU(inplace=True)
        )

        self.lateral_res4 = nn.Sequential(
            nn.Conv3d(in_channels=self.width_per_group * 16 // self.beta_inv,
                      out_channels=self.width_per_group * 16 // self.beta_inv * self.fusion_conv_channel_ratio,
                      kernel_size=(self.fusion_kernel_size, 1, 1),
                      stride=(self.alpha, 1, 1),
                      padding=(self.fusion_kernel_size // 2, 0, 0),
                      bias=False),
            norm_layer(num_features=self.width_per_group * 16 // self.beta_inv * self.fusion_conv_channel_ratio,
                       **({} if norm_kwargs is None else norm_kwargs)),
            nn.ReLU(inplace=True)
        )

        # build slow pathway
        self.slow_conv1 = nn.Conv3d(in_channels=3, out_channels=self.width_per_group,
                                    kernel_size=(1, 7, 7), stride=(1, 2, 2), padding=(0, 3, 3), bias=False)
        self.slow_bn1 = norm_layer(num_features=self.width_per_group,
                                   **({} if norm_kwargs is None else norm_kwargs))
        self.slow_relu = nn.ReLU(inplace=True)
        self.slow_maxpool = nn.MaxPool3d(kernel_size=(1, 3, 3), stride=(1, 2, 2), padding=(0, 1, 1))
        self.slow_res2 = self._make_layer_slow(
            inplanes=self.width_per_group + self.width_per_group // self.out_dim_ratio,
            planes=self.dim_inner,
            num_blocks=layers[0],
            head_conv=1,
            norm_layer=norm_layer,
            norm_kwargs=norm_kwargs,
            layer_name='slow_res2_')
        self.slow_res3 = self._make_layer_slow(
            inplanes=self.width_per_group * 4 + self.width_per_group * 4 // self.out_dim_ratio,
            planes=self.dim_inner * 2,
            num_blocks=layers[1],
            strides=2,
            head_conv=1,
            norm_layer=norm_layer,
            norm_kwargs=norm_kwargs,
            layer_name='slow_res3_')
        self.slow_res4 = self._make_layer_slow(
            inplanes=self.width_per_group * 8 + self.width_per_group * 8 // self.out_dim_ratio,
            planes=self.dim_inner * 4,
            num_blocks=layers[2],
            num_block_temp_kernel_slow=num_block_temp_kernel_slow,
            strides=2,
            head_conv=3,
            norm_layer=norm_layer,
            norm_kwargs=norm_kwargs,
            layer_name='slow_res4_')
        self.slow_res5 = self._make_layer_slow(
            inplanes=self.width_per_group * 16 + self.width_per_group * 16 // self.out_dim_ratio,
            planes=self.dim_inner * 8,
            num_blocks=layers[3],
            strides=2,
            head_conv=3,
            norm_layer=norm_layer,
            norm_kwargs=norm_kwargs,
            layer_name='slow_res5_')

        # build classifier
        self.avg = nn.AdaptiveAvgPool3d(1)
        self.dp = nn.Dropout(p=self.dropout_ratio)
        self.feat_dim = self.width_per_group * 32 // self.beta_inv + self.width_per_group * 32
        self.fc = nn.Linear(in_features=self.feat_dim, out_features=num_classes, bias=True)

    def forward(self, x):
        fast_input = x
        slow_input = x[:, :, ::self.slow_temporal_stride // 2, :, :]

        fast, lateral = self.FastPath(fast_input)
        slow = self.SlowPath(slow_input, lateral)
        x = torch.cat((slow, fast), dim=1)  # bx2304

        if self.feat_ext:
            return x

        x = self.dp(x)
        x = self.fc(x)  # bxnum_classes
        return x

    def SlowPath(self, x, lateral):
        x = self.slow_conv1(x)  # bx64x4x112x112, input is bx3x4x224x224
        x = self.slow_bn1(x)
        x = self.slow_relu(x)
        pool1 = self.slow_maxpool(x)  # bx64x4x56x56
        pool1_lat = torch.cat((pool1, lateral[0]), dim=1)  # bx80x4x56x56

        res2 = self.slow_res2(pool1_lat)  # bx256x4x56x56
        res2_lat = torch.cat((res2, lateral[1]), dim=1)  # bx320x4x56x56

        res3 = self.slow_res3(res2_lat)  # bx512x4x28x28
        res3_lat = torch.cat((res3, lateral[2]), dim=1)  # bx640x4x28x28

        res4 = self.slow_res4(res3_lat)  # bx1024x4x14x14
        res4_lat = torch.cat((res4, lateral[3]), dim=1)  # bx1280x4x14x14

        res5 = self.slow_res5(res4_lat)  # bx2048x4x7x7
        out = self.avg(res5)  # bx2048x1x1x1
        out = out.view(out.shape[0], out.shape[1])  # bx2048
        return out

    def FastPath(self, x):
        lateral = []
        x = self.fast_conv1(x)  # bx8x32x112x112, input is bx3x32x224x224
        x = self.fast_bn1(x)
        x = self.fast_relu(x)
        pool1 = self.fast_maxpool(x)  # bx8x32x56x56
        lateral_p = self.lateral_p1(pool1)  # bx16x4x56x56
        lateral.append(lateral_p)

        res2 = self.fast_res2(pool1)  # bx32x32x56x56
        lateral_res2 = self.lateral_res2(res2)  # bx64x4x56x56
        lateral.append(lateral_res2)

        res3 = self.fast_res3(res2)  # bx64x32x28x28
        lateral_res3 = self.lateral_res3(res3)  # bx128x4x28x28
        lateral.append(lateral_res3)

        res4 = self.fast_res4(res3)  # bx128x32x14x14
        lateral_res4 = self.lateral_res4(res4)  # bx256x4x14x14
        lateral.append(lateral_res4)

        res5 = self.fast_res5(res4)  # bx256x32x7x7
        out = self.avg(res5)  # bx256x1x1x1
        out = out.view(out.shape[0], out.shape[1])  # bx256
        return out, lateral

    def _make_layer_fast(self,
                         inplanes,
                         planes,
                         num_blocks,
                         num_block_temp_kernel_fast=None,
                         block=Bottleneck,
                         strides=1,
                         head_conv=1,
                         norm_layer=BatchNorm3d,
                         norm_kwargs=None,
                         layer_name=''):
        downsample = None
        if strides != 1 or inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv3d(in_channels=inplanes,
                          out_channels=planes * block.expansion,
                          kernel_size=1,
                          stride=(1, strides, strides),
                          bias=False),
                norm_layer(num_features=planes * block.expansion, **({} if norm_kwargs is None else norm_kwargs))
            )

        layers = []
        cnt = 0
        layers.append(block(inplanes=inplanes,
                            planes=planes,
                            strides=strides,
                            downsample=downsample,
                            head_conv=head_conv,
                            layer_name='block%d_' % cnt))
        inplanes = planes * block.expansion
        cnt += 1
        for _ in range(1, num_blocks):
            if num_block_temp_kernel_fast is not None:
                if cnt < num_block_temp_kernel_fast:
                    layers.append(block(inplanes=inplanes,
                                        planes=planes,
                                        head_conv=head_conv,
                                        layer_name='block%d_' % cnt))
                else:
                    layers.append(block(inplanes=inplanes,
                                        planes=planes,
                                        head_conv=1,
                                        layer_name='block%d_' % cnt))
            else:
                layers.append(block(inplanes=inplanes,
                                    planes=planes,
                                    head_conv=head_conv,
                                    layer_name='block%d_' % cnt))
            cnt += 1
        return nn.Sequential(*layers)

    def _make_layer_slow(self,
                         inplanes,
                         planes,
                         num_blocks,
                         num_block_temp_kernel_slow=None,
                         block=Bottleneck,
                         strides=1,
                         head_conv=1,
                         norm_layer=BatchNorm3d,
                         norm_kwargs=None,
                         layer_name=''):
        downsample = None
        if strides != 1 or inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv3d(in_channels=inplanes,
                          out_channels=planes * block.expansion,
                          kernel_size=1,
                          stride=(1, strides, strides),
                          bias=False),
                norm_layer(num_features=planes * block.expansion, **({} if norm_kwargs is None else norm_kwargs))
            )

        layers = []
        cnt = 0
        layers.append(block(inplanes=inplanes,
                            planes=planes,
                            strides=strides,
                            downsample=downsample,
                            head_conv=head_conv,
                            layer_name='block%d_' % cnt))
        inplanes = planes * block.expansion
        cnt += 1
        for _ in range(1, num_blocks):
            if num_block_temp_kernel_slow is not None:
                if cnt < num_block_temp_kernel_slow:
                    layers.append(block(inplanes=inplanes,
                                        planes=planes,
                                        head_conv=head_conv,
                                        layer_name='block%d_' % cnt))
                else:
                    layers.append(block(inplanes=inplanes,
                                        planes=planes,
                                        head_conv=1,
                                        layer_name='block%d_' % cnt))
            else:
                layers.append(block(inplanes=inplanes,
                                    planes=planes,
                                    head_conv=head_conv,
                                    layer_name='block%d_' % cnt))
            cnt += 1
        return nn.Sequential(*layers)


def slowfast_4x16_resnet50_kinetics400(cfg):
    model = SlowFast(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                     layers=[3, 4, 6, 3],
                     pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                     pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                     feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                     num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                     num_crop=cfg.CONFIG.DATA.NUM_CROP,
                     alpha=8,
                     beta_inv=8,
                     fusion_conv_channel_ratio=2,
                     fusion_kernel_size=5,
                     width_per_group=64,
                     num_groups=1,
                     slow_temporal_stride=16,
                     fast_temporal_stride=2,
                     slow_frames=4,
                     fast_frames=32,
                     bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                     partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                     bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('slowfast_4x16_resnet50_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def slowfast_8x8_resnet50_kinetics400(cfg):
    model = SlowFast(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                     layers=[3, 4, 6, 3],
                     pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                     pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                     feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                     num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                     num_crop=cfg.CONFIG.DATA.NUM_CROP,
                     alpha=4,
                     beta_inv=8,
                     fusion_conv_channel_ratio=2,
                     fusion_kernel_size=7,
                     width_per_group=64,
                     num_groups=1,
                     slow_temporal_stride=8,
                     fast_temporal_stride=2,
                     slow_frames=8,
                     fast_frames=32,
                     bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                     partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                     bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('slowfast_8x8_resnet50_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def slowfast_4x16_resnet101_kinetics400(cfg):
    model = SlowFast(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                     layers=[3, 4, 23, 3],
                     pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                     pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                     feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                     num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                     num_crop=cfg.CONFIG.DATA.NUM_CROP,
                     alpha=8,
                     beta_inv=8,
                     fusion_conv_channel_ratio=2,
                     fusion_kernel_size=5,
                     width_per_group=64,
                     num_groups=1,
                     slow_temporal_stride=16,
                     fast_temporal_stride=2,
                     slow_frames=4,
                     fast_frames=32,
                     bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                     partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                     bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)
    return model


def slowfast_8x8_resnet101_kinetics400(cfg):
    model = SlowFast(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                     layers=[3, 4, 23, 3],
                     pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                     pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                     feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                     num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                     num_crop=cfg.CONFIG.DATA.NUM_CROP,
                     alpha=4,
                     beta_inv=8,
                     fusion_conv_channel_ratio=2,
                     fusion_kernel_size=5,
                     width_per_group=64,
                     num_groups=1,
                     slow_temporal_stride=8,
                     fast_temporal_stride=2,
                     slow_frames=8,
                     fast_frames=32,
                     bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                     partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                     bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('slowfast_8x8_resnet101_kinetics400',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def slowfast_16x8_resnet101_kinetics400(cfg):
    model = SlowFast(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                     layers=[3, 4, 23, 3],
                     pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                     pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                     feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                     num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                     num_crop=cfg.CONFIG.DATA.NUM_CROP,
                     alpha=4,
                     beta_inv=8,
                     fusion_conv_channel_ratio=2,
                     fusion_kernel_size=5,
                     width_per_group=64,
                     num_groups=1,
                     slow_temporal_stride=8,
                     fast_temporal_stride=2,
                     slow_frames=16,
                     fast_frames=64,
                     bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                     partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                     bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)
    return model


def slowfast_16x8_resnet101_50_50_kinetics400(cfg):
    model = SlowFast(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                     layers=[3, 4, 23, 3],
                     pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                     pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                     feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                     num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                     num_crop=cfg.CONFIG.DATA.NUM_CROP,
                     num_block_temp_kernel_fast=6,
                     num_block_temp_kernel_slow=6,
                     alpha=4,
                     beta_inv=8,
                     fusion_conv_channel_ratio=2,
                     fusion_kernel_size=5,
                     width_per_group=64,
                     num_groups=1,
                     slow_temporal_stride=8,
                     fast_temporal_stride=2,
                     slow_frames=16,
                     fast_frames=64,
                     bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                     partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                     bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)
    return model


def slowfast_16x8_resnet50_sthsthv2(cfg):
    model = SlowFast(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                     layers=[3, 4, 6, 3],
                     pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                     pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                     feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                     num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                     num_crop=cfg.CONFIG.DATA.NUM_CROP,
                     alpha=4,
                     beta_inv=8,
                     fusion_conv_channel_ratio=2,
                     fusion_kernel_size=7,
                     width_per_group=64,
                     num_groups=1,
                     slow_temporal_stride=8,
                     fast_temporal_stride=2,
                     slow_frames=16,
                     fast_frames=64,
                     bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                     partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                     bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        model.load_state_dict(torch.load(get_model_file('slowfast_16x8_resnet50_sthsthv2',
                                                        tag=cfg.CONFIG.MODEL.PRETRAINED)))
    return model


def slowfast_4x16_resnet50_custom(cfg):
    model = SlowFast(num_classes=cfg.CONFIG.DATA.NUM_CLASSES,
                     layers=[3, 4, 6, 3],
                     pretrained=cfg.CONFIG.MODEL.PRETRAINED,
                     pretrained_base=cfg.CONFIG.MODEL.PRETRAINED_BASE,
                     feat_ext=cfg.CONFIG.INFERENCE.FEAT,
                     num_segment=cfg.CONFIG.DATA.NUM_SEGMENT,
                     num_crop=cfg.CONFIG.DATA.NUM_CROP,
                     alpha=8,
                     beta_inv=8,
                     fusion_conv_channel_ratio=2,
                     fusion_kernel_size=5,
                     width_per_group=64,
                     num_groups=1,
                     slow_temporal_stride=16,
                     fast_temporal_stride=2,
                     slow_frames=4,
                     fast_frames=32,
                     bn_eval=cfg.CONFIG.MODEL.BN_EVAL,
                     partial_bn=cfg.CONFIG.MODEL.PARTIAL_BN,
                     bn_frozen=cfg.CONFIG.MODEL.BN_FROZEN)

    if cfg.CONFIG.MODEL.PRETRAINED:
        from ..model_store import get_model_file
        state_dict = torch.load(get_model_file('slowfast_4x16_resnet50_kinetics400', tag=cfg.CONFIG.MODEL.PRETRAINED))
        for k in list(state_dict.keys()):
            # retain only backbone up to before the classification layer
            if k.startswith('fc'):
                del state_dict[k]

        msg = model.load_state_dict(state_dict, strict=False)
        assert set(msg.missing_keys) == {'fc.weight', 'fc.bias'}
        print("=> initialized from a SlowFast4x16 model pretrained on Kinetcis400 dataset")
    return model



"""Single-shot Multi-box Detector."""
from __future__ import absolute_import

import os
import warnings
import mxnet as mx
from mxnet import autograd
from mxnet.gluon import nn
from mxnet.gluon import HybridBlock
from ...nn.feature import FeatureExpander
from .anchor import SSDAnchorGenerator
from ...nn.predictor import ConvPredictor
from ...nn.coder import MultiPerClassDecoder, NormalizedBoxCenterDecoder
from .vgg_atrous import vgg16_atrous_300, vgg16_atrous_512

__all__ = ['SSD', 'get_ssd', 'custom_ssd']


class SSD(HybridBlock):
    """Single-shot Object Detection Network: https://arxiv.org/abs/1512.02325.

    Parameters
    ----------
    network : string or None
        Name of the base network, if `None` is used, will instantiate the
        base network from `features` directly instead of composing.
    base_size : int
        Base input size, it is speficied so SSD can support dynamic input shapes.
    features : list of str or mxnet.gluon.HybridBlock
        Intermediate features to be extracted or a network with multi-output.
        If `network` is `None`, `features` is expected to be a multi-output network.
    num_filters : list of int
        Number of channels for the appended layers, ignored if `network`is `None`.
    sizes : iterable fo float
        Sizes of anchor boxes, this should be a list of floats, in incremental order.
        The length of `sizes` must be len(layers) + 1. For example, a two stage SSD
        model can have ``sizes = [30, 60, 90]``, and it converts to `[30, 60]` and
        `[60, 90]` for the two stages, respectively. For more details, please refer
        to original paper.
    ratios : iterable of list
        Aspect ratios of anchors in each output layer. Its length must be equals
        to the number of SSD output layers.
    steps : list of int
        Step size of anchor boxes in each output layer.
    classes : iterable of str
        Names of all categories.
    use_1x1_transition : bool
        Whether to use 1x1 convolution as transition layer between attached layers,
        it is effective reducing model capacity.
    use_bn : bool
        Whether to use BatchNorm layer after each attached convolutional layer.
    reduce_ratio : float
        Channel reduce ratio (0, 1) of the transition layer.
    min_depth : int
        Minimum channels for the transition layers.
    global_pool : bool
        Whether to attach a global average pooling layer as the last output layer.
    pretrained : bool or str
        Boolean value controls whether to load the default pretrained weights for model.
        String value represents the hashtag for a certain version of pretrained weights.
    stds : tuple of float, default is (0.1, 0.1, 0.2, 0.2)
        Std values to be divided/multiplied to box encoded values.
    nms_thresh : float, default is 0.45.
        Non-maximum suppression threshold. You can specify < 0 or > 1 to disable NMS.
    nms_topk : int, default is 400
        Apply NMS to top k detection results, use -1 to disable so that every Detection
         result is used in NMS.
    post_nms : int, default is 100
        Only return top `post_nms` detection results, the rest is discarded. The number is
        based on COCO dataset which has maximum 100 objects per image. You can adjust this
        number if expecting more objects. You can use -1 to return all detections.
    anchor_alloc_size : tuple of int, default is (128, 128)
        For advanced users. Define `anchor_alloc_size` to generate large enough anchor
        maps, which will later saved in parameters. During inference, we support arbitrary
        input image by cropping corresponding area of the anchor map. This allow us
        to export to symbol so we can run it in c++, scalar, etc.
    ctx : mx.Context
        Network context.
    norm_layer : object
        Normalization layer used (default: :class:`mxnet.gluon.nn.BatchNorm`)
        Can be :class:`mxnet.gluon.nn.BatchNorm` or :class:`mxnet.gluon.contrib.nn.SyncBatchNorm`.
        This will only apply to base networks that has `norm_layer` specified, will ignore if the
        base network (e.g. VGG) don't accept this argument.
    norm_kwargs : dict
        Additional `norm_layer` arguments, for example `num_devices=4`
        for :class:`mxnet.gluon.contrib.nn.SyncBatchNorm`.
    root : str
        The root path for model storage, default is '~/.mxnet/models'
    minimal_opset : bool
        We sometimes add special operators to accelerate training/inference, however, for exporting
        to third party compilers we want to utilize most widely used operators.
        If `minimal_opset` is `True`, the network will use a minimal set of operators good
        for e.g., `TVM`.
    predictor_kernel: tuple of int. default is (3,3)
        Dimension of predictor kernel
    predictor_pad: tuple of int. default is (1,1)
        Padding of the predictor kenrel conv.
    anchor_generator: default is SSDAnchorGenerator
        Anchor Generator to be used. The default it SSDAnchorGenerator corresponding
        to SSD published article. This argument can be used for other custom
        anchor generators. Like LiteAnchorGenerator.

    """
    def __init__(self, network, base_size, features, num_filters, sizes, ratios,
                 steps, classes, use_1x1_transition=True, use_bn=True,
                 reduce_ratio=1.0, min_depth=128, global_pool=False, pretrained=False,
                 stds=(0.1, 0.1, 0.2, 0.2), nms_thresh=0.45, nms_topk=400, post_nms=100,
                 anchor_alloc_size=128, ctx=mx.cpu(),
                 norm_layer=nn.BatchNorm, norm_kwargs=None,
                 root=os.path.join('~', '.mxnet', 'models'), minimal_opset=False,
                 predictors_kernel=(3, 3), predictors_pad=(1, 1),
                 anchor_generator=SSDAnchorGenerator, **kwargs):
        super(SSD, self).__init__(**kwargs)
        if norm_kwargs is None:
            norm_kwargs = {}
        if network is None:
            num_layers = len(ratios)
        else:
            num_layers = len(features) + len(num_filters) + int(global_pool)
        assert len(sizes) == num_layers + 1
        sizes = list(zip(sizes[:-1], sizes[1:]))
        assert isinstance(ratios, list), "Must provide ratios as list or list of list"
        if not isinstance(ratios[0], (tuple, list)):
            ratios = ratios * num_layers  # propagate to all layers if use same ratio
        assert num_layers == len(sizes) == len(ratios), \
            "Mismatched (number of layers) vs (sizes) vs (ratios): {}, {}, {}".format(
                num_layers, len(sizes), len(ratios))
        assert num_layers > 0, "SSD require at least one layer, suggest multiple."
        self._num_layers = num_layers
        self.classes = classes
        self.nms_thresh = nms_thresh
        self.nms_topk = nms_topk
        self.post_nms = post_nms

        with self.name_scope():
            if network is None:
                # use fine-grained manually designed block as features
                try:
                    self.features = features(pretrained=pretrained, ctx=ctx, root=root,
                                             norm_layer=norm_layer, norm_kwargs=norm_kwargs)
                except TypeError:
                    self.features = features(pretrained=pretrained, ctx=ctx, root=root)
            else:
                try:
                    self.features = FeatureExpander(
                        network=network, outputs=features, num_filters=num_filters,
                        use_1x1_transition=use_1x1_transition,
                        use_bn=use_bn, reduce_ratio=reduce_ratio, min_depth=min_depth,
                        global_pool=global_pool, pretrained=pretrained, ctx=ctx,
                        norm_layer=norm_layer, norm_kwargs=norm_kwargs, root=root)
                except TypeError:
                    self.features = FeatureExpander(
                        network=network, outputs=features, num_filters=num_filters,
                        use_1x1_transition=use_1x1_transition,
                        use_bn=use_bn, reduce_ratio=reduce_ratio, min_depth=min_depth,
                        global_pool=global_pool, pretrained=pretrained, ctx=ctx, root=root)
            self.class_predictors = nn.HybridSequential()
            self.box_predictors = nn.HybridSequential()
            self.anchor_generators = nn.HybridSequential()
            asz = anchor_alloc_size
            im_size = (base_size, base_size)
            for i, s, r, st in zip(range(num_layers), sizes, ratios, steps):
                branch_anchor_generator = anchor_generator(i, im_size, s, r, st, (asz, asz))
                self.anchor_generators.add(branch_anchor_generator)
                asz = max(asz // 2, 16)  # pre-compute larger than 16x16 anchor map
                num_anchors = branch_anchor_generator.num_depth
                self.class_predictors.add(ConvPredictor(num_anchors * (len(self.classes) + 1),
                                                        kernel=predictors_kernel,
                                                        pad=predictors_pad))
                self.box_predictors.add(ConvPredictor(num_anchors * 4,
                                                      kernel=predictors_kernel,
                                                      pad=predictors_pad))
            self.bbox_decoder = NormalizedBoxCenterDecoder(stds, minimal_opset=minimal_opset)
            self.cls_decoder = MultiPerClassDecoder(len(self.classes) + 1, thresh=0.01)

    @property
    def num_classes(self):
        """Return number of foreground classes.

        Returns
        -------
        int
            Number of foreground classes

        """
        return len(self.classes)

    def set_nms(self, nms_thresh=0.45, nms_topk=400, post_nms=100):
        """Set non-maximum suppression parameters.

        Parameters
        ----------
        nms_thresh : float, default is 0.45.
            Non-maximum suppression threshold. You can specify < 0 or > 1 to disable NMS.
        nms_topk : int, default is 400
            Apply NMS to top k detection results, use -1 to disable so that every Detection
             result is used in NMS.
        post_nms : int, default is 100
            Only return top `post_nms` detection results, the rest is discarded. The number is
            based on COCO dataset which has maximum 100 objects per image. You can adjust this
            number if expecting more objects. You can use -1 to return all detections.

        Returns
        -------
        None

        """
        self._clear_cached_op()
        self.nms_thresh = nms_thresh
        self.nms_topk = nms_topk
        self.post_nms = post_nms

    # pylint: disable=arguments-differ
    def hybrid_forward(self, F, x):
        """Hybrid forward"""
        features = self.features(x)
        cls_preds = [F.flatten(F.transpose(cp(feat), (0, 2, 3, 1)))
                     for feat, cp in zip(features, self.class_predictors)]
        box_preds = [F.flatten(F.transpose(bp(feat), (0, 2, 3, 1)))
                     for feat, bp in zip(features, self.box_predictors)]
        anchors = [F.reshape(ag(feat), shape=(1, -1))
                   for feat, ag in zip(features, self.anchor_generators)]
        cls_preds = F.concat(*cls_preds, dim=1).reshape((0, -1, self.num_classes + 1))
        box_preds = F.concat(*box_preds, dim=1).reshape((0, -1, 4))
        anchors = F.concat(*anchors, dim=1).reshape((1, -1, 4))
        if autograd.is_training():
            return [cls_preds, box_preds, anchors]
        bboxes = self.bbox_decoder(box_preds, anchors)
        cls_ids, scores = self.cls_decoder(F.softmax(cls_preds, axis=-1))
        results = []
        for i in range(self.num_classes):
            cls_id = cls_ids.slice_axis(axis=-1, begin=i, end=i+1)
            score = scores.slice_axis(axis=-1, begin=i, end=i+1)
            # per class results
            per_result = F.concat(*[cls_id, score, bboxes], dim=-1)
            results.append(per_result)
        result = F.concat(*results, dim=1)
        if self.nms_thresh > 0 and self.nms_thresh < 1:
            result = F.contrib.box_nms(
                result, overlap_thresh=self.nms_thresh, topk=self.nms_topk, valid_thresh=0.01,
                id_index=0, score_index=1, coord_start=2, force_suppress=False)
            if self.post_nms > 0:
                result = result.slice_axis(axis=1, begin=0, end=self.post_nms)
        ids = F.slice_axis(result, axis=2, begin=0, end=1)
        scores = F.slice_axis(result, axis=2, begin=1, end=2)
        bboxes = F.slice_axis(result, axis=2, begin=2, end=6)
        return ids, scores, bboxes

    def reset_class(self, classes, reuse_weights=None):
        """Reset class categories and class predictors.

        Parameters
        ----------
        classes : iterable of str
            The new categories. ['apple', 'orange'] for example.
        reuse_weights : dict
            A {new_integer : old_integer} or mapping dict or {new_name : old_name} mapping dict,
            or a list of [name0, name1,...] if class names don't change.
            This allows the new predictor to reuse the
            previously trained weights specified.

        Example
        -------
        >>> net = gluoncv.model_zoo.get_model('ssd_512_resnet50_v1_voc', pretrained=True)
        >>> # use direct name to name mapping to reuse weights
        >>> net.reset_class(classes=['person'], reuse_weights={'person':'person'})
        >>> # or use interger mapping, person is the 14th category in VOC
        >>> net.reset_class(classes=['person'], reuse_weights={0:14})
        >>> # you can even mix them
        >>> net.reset_class(classes=['person'], reuse_weights={'person':14})
        >>> # or use a list of string if class name don't change
        >>> net.reset_class(classes=['person'], reuse_weights=['person'])

        """
        self._clear_cached_op()
        old_classes = self.classes
        self.classes = classes
        # trying to reuse weights by mapping old and new classes
        if isinstance(reuse_weights, (dict, list)):
            if isinstance(reuse_weights, dict):
                # trying to replace str with indices
                new_keys = []
                new_vals = []
                for k, v in reuse_weights.items():
                    if isinstance(v, str):
                        try:
                            new_vals.append(old_classes.index(v))  # raise ValueError if not found
                        except ValueError:
                            raise ValueError(
                                "{} not found in old class names {}".format(v, old_classes))
                    else:
                        if v < 0 or v >= len(old_classes):
                            raise ValueError(
                                "Index {} out of bounds for old class names".format(v))
                        new_vals.append(v)
                    if isinstance(k, str):
                        try:
                            new_keys.append(self.classes.index(k))  # raise ValueError if not found
                        except ValueError:
                            raise ValueError(
                                "{} not found in new class names {}".format(k, self.classes))
                    else:
                        if k < 0 or k >= len(self.classes):
                            raise ValueError(
                                "Index {} out of bounds for new class names".format(k))
                        new_keys.append(k)
                reuse_weights = dict(zip(new_keys, new_vals))
            else:
                new_map = {}
                for x in reuse_weights:
                    try:
                        new_idx = self.classes.index(x)
                        old_idx = old_classes.index(x)
                        new_map[new_idx] = old_idx
                    except ValueError:
                        warnings.warn("{} not found in old: {} or new class names: {}".format(
                            x, old_classes, self.classes))
                reuse_weights = new_map
        # replace class predictors
        with self.name_scope():
            class_predictors = nn.HybridSequential(prefix=self.class_predictors.prefix)
            for i, ag in zip(range(len(self.class_predictors)), self.anchor_generators):
                # Re-use the same prefix and ctx_list as used by the current ConvPredictor
                prefix = self.class_predictors[i].prefix
                old_pred = self.class_predictors[i].predictor
                ctx = list(old_pred.params.values())[0].list_ctx()
                # to avoid deferred init, number of in_channels must be defined
                in_channels = list(old_pred.params.values())[0].shape[1]
                new_cp = ConvPredictor(ag.num_depth * (self.num_classes + 1),
                                       in_channels=in_channels, prefix=prefix)
                new_cp.collect_params().initialize(ctx=ctx)
                if reuse_weights:
                    assert isinstance(reuse_weights, dict)
                    for old_params, new_params in zip(old_pred.params.values(),
                                                      new_cp.predictor.params.values()):
                        old_data = old_params.data()
                        new_data = new_params.data()

                        for k, v in reuse_weights.items():
                            if k >= len(self.classes) or v >= len(old_classes):
                                warnings.warn("reuse mapping {}/{} -> {}/{} out of range".format(
                                    k, self.classes, v, old_classes))
                                continue
                            # always increment k and v (background is always the 0th)
                            new_data[k+1::len(self.classes)+1] = old_data[v+1::len(old_classes)+1]
                        # reuse background weights as well
                        new_data[0::len(self.classes)+1] = old_data[0::len(old_classes)+1]
                        # set data to new conv layers
                        new_params.set_data(new_data)
                class_predictors.add(new_cp)
            self.class_predictors = class_predictors
            self.cls_decoder = MultiPerClassDecoder(len(self.classes) + 1, thresh=0.01)

def get_ssd(name, base_size, features, filters, sizes, ratios, steps, classes,
            dataset, pretrained=False, pretrained_base=True, ctx=mx.cpu(),
            root=os.path.join('~', '.mxnet', 'models'),
            anchor_generator=SSDAnchorGenerator, **kwargs):
    """Get SSD models.

    Parameters
    ----------
    name : str or None
        Model name, if `None` is used, you must specify `features` to be a `HybridBlock`.
    base_size : int
        Base image size for training, this is fixed once training is assigned.
        A fixed base size still allows you to have variable input size during test.
    features : iterable of str or `HybridBlock`
        List of network internal output names, in order to specify which layers are
        used for predicting bbox values.
        If `name` is `None`, `features` must be a `HybridBlock` which generate multiple
        outputs for prediction.
    filters : iterable of float or None
        List of convolution layer channels which is going to be appended to the base
        network feature extractor. If `name` is `None`, this is ignored.
    sizes : iterable fo float
        Sizes of anchor boxes, this should be a list of floats, in incremental order.
        The length of `sizes` must be len(layers) + 1. For example, a two stage SSD
        model can have ``sizes = [30, 60, 90]``, and it converts to `[30, 60]` and
        `[60, 90]` for the two stages, respectively. For more details, please refer
        to original paper.
    ratios : iterable of list
        Aspect ratios of anchors in each output layer. Its length must be equals
        to the number of SSD output layers.
    steps : list of int
        Step size of anchor boxes in each output layer.
    classes : iterable of str
        Names of categories.
    dataset : str
        Name of dataset. This is used to identify model name because models trained on
        different datasets are going to be very different.
    pretrained : bool or str
        Boolean value controls whether to load the default pretrained weights for model.
        String value represents the hashtag for a certain version of pretrained weights.
    pretrained_base : bool or str, optional, default is True
        Load pretrained base network, the extra layers are randomized. Note that
        if pretrained is `True`, this has no effect.
    ctx : mxnet.Context
        Context such as mx.cpu(), mx.gpu(0).
    root : str
        Model weights storing path.
    norm_layer : object
        Normalization layer used (default: :class:`mxnet.gluon.nn.BatchNorm`)
        Can be :class:`mxnet.gluon.nn.BatchNorm` or :class:`mxnet.gluon.contrib.nn.SyncBatchNorm`.
    norm_kwargs : dict
        Additional `norm_layer` arguments, for example `num_devices=4`
        for :class:`mxnet.gluon.contrib.nn.SyncBatchNorm`.

    Returns
    -------
    HybridBlock
        A SSD detection network.
    """
    pretrained_base = False if pretrained else pretrained_base
    base_name = None if callable(features) else name
    net = SSD(base_name, base_size, features, filters, sizes, ratios, steps,
              pretrained=pretrained_base, classes=classes, ctx=ctx, root=root,
              minimal_opset=pretrained, anchor_generator=anchor_generator, **kwargs)
    if pretrained:
        from ..model_store import get_model_file
        full_name = '_'.join(('ssd', str(base_size), name, dataset))
        net.load_parameters(get_model_file(full_name, tag=pretrained, root=root), ctx=ctx)
    return net

def custom_ssd(base_network_name, base_size, filters, sizes, ratios, steps,
               classes, dataset, pretrained_base, **kwargs):
    """Custom SSD models.
    """
    if base_network_name == 'vgg16_atrous' and base_size == 300:
        features = vgg16_atrous_300
    elif base_network_name == 'vgg16_atrous' and base_size == 512:
        features = vgg16_atrous_512
    elif base_network_name == 'resnet18_v1' and base_size == 512:
        features = ['stage3_activation1', 'stage4_activation1']
    elif base_network_name == 'resnet50_v1' and base_size == 512:
        features = ['stage3_activation5', 'stage4_activation2']
    elif base_network_name == 'resnet101_v2' and base_size == 512:
        features = ['stage3_activation22', 'stage4_activation2']
    elif base_network_name == 'resnet152_v2' and base_size == 512:
        features = ['stage2_activation7', 'stage3_activation35', 'stage4_activation2']
    elif base_network_name == 'mobilenet1.0' and base_size == 512:
        features = ['relu22_fwd', 'relu26_fwd']
    elif base_network_name == 'mobilenet1.0' and base_size == 300:
        features = ['relu22_fwd', 'relu26_fwd']
    elif base_network_name == 'mobilenet0.25' and base_size == 300:
        features = ['relu22_fwd', 'relu26_fwd']
    else:
        raise NotImplementedError('Unsupported network', base_network_name)

    net = get_ssd(name=base_network_name,
                  base_size=base_size,
                  features=features,
                  filters=filters,
                  sizes=sizes,
                  ratios=ratios,
                  steps=steps,
                  classes=classes,
                  dataset=dataset,
                  pretrained_base=pretrained_base,
                  **kwargs)

    return net



"""GluonCV Model Zoo"""
# pylint: disable=wildcard-import
from .model_zoo import get_model, get_model_list
from .model_store import pretrained_model_list
from .rcnn.faster_rcnn import *
from .rcnn.mask_rcnn import *
from .ssd import *
from .yolo import *
from .cifarresnet import *
from .cifarwideresnet import *
from .fcn import *
from .pspnet import *
from .deeplabv3 import *
from .deeplabv3_plus import *
from .deeplabv3b_plus import *
from . import segbase
from .resnetv1b import *
from .se_resnet import *
from .nasnet import *
from .simple_pose.simple_pose_resnet import *
from .simple_pose.mobile_pose import *
from .action_recognition import *
from .wideresnet import *

from .resnest import *
from .resnext import *
from .alexnet import *
from .densenet import *
from .googlenet import *
from .inception import *
from .xception import *
from .resnet import *
from .squeezenet import *
from .vgg import *
from .mobilenet import *
from .residual_attentionnet import *
from .center_net import *
from .hrnet import *
from .siamrpn import *
from .fastscnn import *
from .monodepthv2 import *
from .smot import *



import mxnet as mx
from mxnet import nd, autograd, gluon
from mxnet.gluon import nn, Block
from gluoncv.utils.parallel import DataParallelModel, DataParallelCriterion

def test_data_parallel():
    # test gluon.contrib.parallel.DataParallelModel
    net = nn.HybridSequential()
    with net.name_scope():
        net.add(nn.Conv2D(in_channels=1, channels=5, kernel_size=5))
        net.add(nn.Activation('relu'))
        net.add(nn.MaxPool2D(pool_size=2, strides=2))
        net.add(nn.Conv2D(in_channels=5, channels=5, kernel_size=5))
        net.add(nn.Activation('relu'))
        net.add(nn.MaxPool2D(pool_size=2, strides=2))
        # The Flatten layer collapses all axis, except the first one, into one axis.
        net.add(nn.Flatten())
        net.add(nn.Dense(8,in_units=80))
        net.add(nn.Activation('relu'))
        net.add(nn.Dense(10, in_units=8))

    net.collect_params().initialize()
    criterion = gluon.loss.SoftmaxCELoss(axis=1)

    def test_net_sync(net, criterion, sync, nDevices):
        ctx_list = [mx.cpu(0) for i in range(nDevices)]
        net = DataParallelModel(net, ctx_list, sync=sync)
        criterion = DataParallelCriterion(criterion, ctx_list, sync=sync)
        iters = 10
        bs = 2
        # train mode
        for i in range(iters):
            x = mx.random.uniform(shape=(bs, 1, 28, 28))
            t = nd.ones(shape=(bs))
            with autograd.record():
                y = net(x)
                loss = criterion(y, t)
                autograd.backward(loss)
        # evaluation mode
        for i in range(iters):
            x = mx.random.uniform(shape=(bs, 1, 28, 28))
            y = net(x)
        nd.waitall()

    # test_net_sync(net, criterion, True, 1)
    test_net_sync(net, criterion, True, 2)
    # test_net_sync(net, criterion, False, 1)
    test_net_sync(net, criterion, False, 2)


if __name__ == "__main__":
    import nose
    nose.runmodule()



"""Faster RCNN default config"""
# pylint: disable=unused-variable,missing-function-docstring,bad-whitespace,missing-class-docstring
from typing import Union, Tuple, Any
from autocfg import dataclass, field

@dataclass
class FasterRCNN:
    # Backbone network.
    backbone : str = 'resnet50_v1b'  # base feature network
    # Final R-CNN non-maximum suppression threshold. You can specify < 0 or > 1 to disable NMS.
    nms_thresh : float = 0.5
    # Apply R-CNN NMS to top k detection results, use -1 to disable so that every Detection
    # result is used in NMS.
    nms_topk : int = -1
    # ROI pooling mode. Currently support 'pool' and 'align'.
    roi_mode : str = 'align'
    # (height, width) of the ROI region.
    roi_size : Union[Tuple, list] = (7, 7)
    # Feature map stride with respect to original image.
    # This is usually the ratio between original image size and feature map size.
    # For FPN, use a tuple of ints.
    strides : Union[Tuple, list] = (4, 8, 16, 32, 64)
    # Clip bounding box prediction to to prevent exponentiation from overflowing.
    clip : float = 4.14

    # Anchors generation
    # ------------------
    # The width(and height) of reference anchor box.
    anchor_base_size : int = 16
    # The areas of anchor boxes.
    # We use the following form to compute the shapes of anchors:
    # .. math::
    # width_{anchor} = size_{base} \times scale \times \sqrt{ 1 / ratio}
    # height_{anchor} = size_{base} \times scale \times \sqrt{ratio}
    anchor_aspect_ratio : Union[Tuple, list] = (0.5, 1, 2)
    # The aspect ratios of anchor boxes. We expect it to be a list or tuple.
    anchor_scales : Union[Tuple, list] = (2, 4, 8, 16, 32)

    # Allocate size for the anchor boxes as (H, W).
    # Usually we generate enough anchors for large feature map, e.g. 128x128.
    # Later in inference we can have variable input sizes,
    # at which time we can crop corresponding anchors from this large
    # anchor map so we can skip re-generating anchors for each input.
    anchor_alloc_size : Union[Tuple, list] = (384, 384)

    # number of channels used in RPN convolutional layers.
    rpn_channel : int = 256
    # IOU threshold for NMS. It is used to remove overlapping proposals.
    rpn_nms_thresh : float = 0.7
    # Maximum ground-truth number for each example. This is only an upper bound, not
    # necessarily very precise. However, using a very big number may impact the training speed.
    max_num_gt : int = 100
    # Gluon normalization layer to use. Default is none which will use frozen
    # batch normalization layer.
    norm_layer : Any = None

    # FPN Options
    # -----------
    # Whether to use FPN.
    use_fpn : bool = True
    # Number of filters for FPN output layers.
    num_fpn_filters : int = 256

    # Number of convolution layers to use in box head if batch normalization is not frozen.
    num_box_head_conv : int = 4
    # Number of filters for convolution layers in box head.
    # Only applicable if batch normalization is not frozen.
    num_box_head_conv_filters : int = 256
    # Number of hidden units for the last fully connected layer in box head.
    num_box_head_dense_filters : int = 1024

    # Input image short side size.
    image_short : int = 800
    # Maximum size of input image long side.
    image_max_size : int = 1333

    # Whether to enable custom model.
    # custom_model = True
    # Whether to use automatic mixed precision
    amp : bool = False
    # Whether to allocate memory statically.
    static_alloc : bool = False
    # whether apply transfer learning from pre-trained models, if True, override other net structures
    transfer : Union[str, None] = 'faster_rcnn_fpn_resnet50_v1b_coco'


@dataclass
class TrainCfg:
    # Whether load the imagenet pre-trained base
    pretrained_base : bool = True
    # Batch size during training
    batch_size : int = 1
    # starting epoch
    start_epoch : int = 0
    # total epoch for training
    epochs : int = 10

    # Solver
    # ------
    # Learning rate.
    lr : float = 0.01
    # Decay rate of learning rate.
    lr_decay : float = 0.1
    # Epochs at which learning rate decays
    lr_decay_epoch : Union[Tuple, list] = (20, 24)
    # Learning rate scheduler mode. options are step, poly and cosine
    lr_mode : str = 'step'
    # Number of iterations for warmup.
    lr_warmup : int = 500
    # Starging lr warmup factor.
    lr_warmup_factor : float = 1. / 3.
    # Momentum
    momentum : float = 0.9
    # Weight decay
    wd : float = 1e-4

    # RPN options
    # -----------
    # Filter top proposals before NMS in training of RPN.
    rpn_train_pre_nms : int = 12000
    # Return top proposal results after NMS in training of RPN.
    # Will be set to rpn_train_pre_nms if it is larger than rpn_train_pre_nms.
    rpn_train_post_nms : int = 2000
    # RPN box regression transition point from L1 to L2 loss.
    # Set to 0.0 to make the loss simply L1.
    rpn_smoothl1_rho : float = 0.001
    # Proposals whose size is smaller than ``min_size`` will be discarded.
    rpn_min_size : int = 1

    # R-CNN options
    # -------------
    # Number of samples for RPN targets.
    rcnn_num_samples : int = 512
    # Anchor with IOU larger than ``rcnn_pos_iou_thresh`` is regarded as positive samples.
    rcnn_pos_iou_thresh : float = 0.5
    # ``rcnn_pos_iou_thresh`` defines how many positive samples
    # (``rcnn_pos_iou_thresh * num_sample``) is to be sampled.
    rcnn_pos_ratio : float = 0.25
    # R-CNN box regression transition point from L1 to L2 loss.
    # Set to 0.0 to make the loss simply L1.
    rcnn_smoothl1_rho : float = 0.001

    # Misc
    # ----
    # log interval in terms of iterations
    log_interval : int = 100
    # Random seed to be fixed.
    seed : int = 233
    # Whether to enable verbose logging
    verbose : bool = False
    # Whether to enable mixup training
    mixup : bool = False
    # If mixup is enable, disable mixup after ```no_mixup_epochs```.
    no_mixup_epochs : int = 20
    # Number of threads for executor for scheduling ops.
    # More threads may incur higher GPU memory footprint,
    # but may speed up throughput. Note that when horovod is used,
    # it is set to 1.
    executor_threads : int = 4


@dataclass
class ValidCfg:
    # Batch size during training
    batch_size : int = 1
    # Filter top proposals before NMS in testing of RPN.
    rpn_test_pre_nms : int = 6000
    # Return top proposal results after NMS in testing of RPN.
    # Will be set to rpn_test_pre_nms if it is larger than rpn_test_pre_nms.
    rpn_test_post_nms : int = 1000
    # Epoch interval for validation
    val_interval : int = 1
    # metric, 'voc', 'voc07'
    metric : str = 'voc07'
    # iou_thresh for VOC type metrics
    iou_thresh : float = 0.5

@dataclass
class FasterRCNNCfg:
    faster_rcnn : FasterRCNN = field(default_factory=FasterRCNN)
    train : TrainCfg = field(default_factory=TrainCfg)
    valid : ValidCfg = field(default_factory=ValidCfg)
    # Dataset name. eg. 'coco', 'voc', 'voc_tiny'
    dataset : str = 'voc_tiny'
    # Path of the directory where the dataset is located.
    dataset_root : str = '~/.mxnet/datasets/'
    # Training with GPUs, you can specify (1,3) for example.
    gpus : Union[Tuple, list] = (0, 1, 2, 3)
    # Resume from previously saved parameters if not None.
    # For example, you can resume from ./faster_rcnn_xxx_0123.params.
    resume : str = ''
    # Saving parameter prefix
    save_prefix : str = ''
    # Saving parameters epoch interval, best model will always be saved.
    save_interval : int = 1
    # Use MXNet Horovod for distributed training. Must be run with OpenMPI.
    horovod : bool = False
    # Number of data workers, you can use larger number to accelerate data loading,
    # if your CPU and GPUs are powerful.
    num_workers : int = 16
    # KV store options. local, device, nccl, dist_sync, dist_device_sync,
    # dist_async are available.
    kv_store : str = 'nccl'
    # Whether to disable hybridize the model. Memory usage and speed will decrese.
    disable_hybridization : bool = False



"""Data parallel task for Faster RCNN Model."""

from mxnet import autograd

from gluoncv.utils.parallel import Parallelizable


class ForwardBackwardTask(Parallelizable):
    """ Faster R-CNN training task that can be scheduled concurrently using Parallel.
    Parameters
    ----------
    net : gluon.HybridBlock
        Faster R-CNN network.
    optimizer : gluon.Trainer
        Optimizer for the training.
    rpn_cls_loss : gluon.loss
        RPN box classification loss.
    rpn_box_loss : gluon.loss
        RPN box regression loss.
    rcnn_cls_loss : gluon.loss
        R-CNN box head classification loss.
    rcnn_box_loss : gluon.loss
        R-CNN box head regression loss.
    mix_ratio : float
        Object detection mixup ratio.
    amp_enabled : bool
        Whether to enable Automatic Mixed Precision.
    """

    def __init__(self, net, optimizer, rpn_cls_loss, rpn_box_loss, rcnn_cls_loss, rcnn_box_loss,
                 mix_ratio, amp_enabled):
        super(ForwardBackwardTask, self).__init__()
        self.net = net
        ctx = list(net.collect_params().values())[0].list_ctx()
        self.net.target_generator.collect_params().reset_ctx(ctx)
        self._optimizer = optimizer
        self.rpn_cls_loss = rpn_cls_loss
        self.rpn_box_loss = rpn_box_loss
        self.rcnn_cls_loss = rcnn_cls_loss
        self.rcnn_box_loss = rcnn_box_loss
        self.mix_ratio = mix_ratio
        self.amp_enabled = amp_enabled

    def forward_backward(self, x):
        data, label, rpn_cls_targets, rpn_box_targets, rpn_box_masks = x
        with autograd.record():
            gt_label = label[:, :, 4:5]
            gt_box = label[:, :, :4]
            cls_pred, box_pred, _, _, _Z, rpn_score, rpn_box, _, cls_targets, \
                box_targets, box_masks, _ = self.net(data, gt_box, gt_label)
            # losses of rpn
            rpn_score = rpn_score.squeeze(axis=-1)
            num_rpn_pos = (rpn_cls_targets >= 0).sum()
            rpn_loss1 = self.rpn_cls_loss(rpn_score, rpn_cls_targets,
                                          rpn_cls_targets >= 0) * rpn_cls_targets.size / num_rpn_pos
            rpn_loss2 = self.rpn_box_loss(rpn_box, rpn_box_targets,
                                          rpn_box_masks) * rpn_box.size / num_rpn_pos
            # rpn overall loss, use sum rather than average
            rpn_loss = rpn_loss1 + rpn_loss2
            # losses of rcnn
            num_rcnn_pos = (cls_targets >= 0).sum()
            rcnn_loss1 = self.rcnn_cls_loss(
                cls_pred, cls_targets, cls_targets.expand_dims(-1) >= 0) * cls_targets.size / \
                         num_rcnn_pos
            rcnn_loss2 = self.rcnn_box_loss(box_pred, box_targets, box_masks) * box_pred.size / \
                         num_rcnn_pos
            rcnn_loss = rcnn_loss1 + rcnn_loss2
            # overall losses
            total_loss = rpn_loss.sum() * self.mix_ratio + rcnn_loss.sum() * self.mix_ratio

            rpn_loss1_metric = rpn_loss1.mean() * self.mix_ratio
            rpn_loss2_metric = rpn_loss2.mean() * self.mix_ratio
            rcnn_loss1_metric = rcnn_loss1.mean() * self.mix_ratio
            rcnn_loss2_metric = rcnn_loss2.mean() * self.mix_ratio
            rpn_acc_metric = [[rpn_cls_targets, rpn_cls_targets >= 0], [rpn_score]]
            rpn_l1_loss_metric = [[rpn_box_targets, rpn_box_masks], [rpn_box]]
            rcnn_acc_metric = [[cls_targets], [cls_pred]]
            rcnn_l1_loss_metric = [[box_targets, box_masks], [box_pred]]

            if self.amp_enabled:
                from mxnet.contrib import amp
                with amp.scale_loss(total_loss, self._optimizer) as scaled_losses:
                    autograd.backward(scaled_losses)
            else:
                total_loss.backward()

        return rpn_loss1_metric, rpn_loss2_metric, rcnn_loss1_metric, rcnn_loss2_metric, \
               rpn_acc_metric, rpn_l1_loss_metric, rcnn_acc_metric, rcnn_l1_loss_metric
